"""Methods for clean null or NaN values in a dataset"""
from dataclasses import dataclass
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Union

import numpy as np
import pandas as pd

from evidently import ColumnMapping


def replace_infinity_values_to_nan(dataframe: pd.DataFrame) -> pd.DataFrame:
    #   document somewhere, that all analyzers are mutators, i.e. they will change
    #   the dataframe, like here: replace inf and nan values.
    dataframe.replace([np.inf, -np.inf], np.nan, inplace=True)
    return dataframe


@dataclass
class DatasetUtilityColumns:
    date: Optional[str]
    id_column: Optional[str]
    target: Optional[str]
    prediction: Optional[Union[str, Sequence[str]]]

    def as_dict(
        self,
    ) -> Dict[str, Union[Optional[str], Optional[Union[str, Sequence[str]]]]]:
        return {
            "date": self.date,
            "id": self.id_column,
            "target": self.target,
            "prediction": self.prediction,
        }


@dataclass
class DatasetColumns:
    utility_columns: DatasetUtilityColumns
    target_type: Optional[str]
    num_feature_names: List[str]
    cat_feature_names: List[str]
    text_feature_names: List[str]
    datetime_feature_names: List[str]
    target_names: Optional[List[str]]
    task: Optional[str]

    def as_dict(self) -> Dict[str, Union[str, Optional[List[str]], Dict]]:
        return {
            "utility_columns": self.utility_columns.as_dict(),
            "cat_feature_names": self.cat_feature_names,
            "num_feature_names": self.num_feature_names,
            "datetime_feature_names": self.datetime_feature_names,
            "target_names": self.target_names,
            "text_feature_names": self.text_feature_names,
        }

    def get_all_features_list(self, cat_before_num: bool = True, include_datetime_feature: bool = False) -> List[str]:
        """List all features names.

        By default, returns cat features than num features and du not return other.

        If you want to change the order - set  `cat_before_num` to False.

        If you want to add date time columns - set `include_datetime_feature` to True.
        """
        if cat_before_num:
            result = self.cat_feature_names + self.num_feature_names + self.text_feature_names

        else:
            result = self.num_feature_names + self.cat_feature_names + self.text_feature_names

        if include_datetime_feature and self.datetime_feature_names:
            result += self.datetime_feature_names

        return result

    def get_all_columns_list(self, skip_id_column: bool = False, skip_text_columns: bool = False) -> List[str]:
        """List all columns."""
        result: List[str] = self.cat_feature_names + self.num_feature_names

        if not skip_text_columns:
            result.extend(self.text_feature_names)

        result.extend(
            [
                name
                for name in (
                    self.utility_columns.id_column if not skip_id_column else None,
                    self.utility_columns.date,
                    self.utility_columns.target,
                    self.utility_columns.prediction,
                )
                if name is not None and isinstance(name, str)
            ]
        )
        return result

    def get_features_len(self, include_time_columns: bool = False) -> int:
        """How mane feature do we have. It is useful for pagination in widgets.

        By default, we sum category nad numeric features.

        If you want to include date time columns - set `include_datetime_feature` to True.
        """
        if include_time_columns and self.datetime_feature_names:
            len_time_columns = len(self.datetime_feature_names)

        else:
            len_time_columns = 0

        return (
            len(self.num_feature_names) + len(self.cat_feature_names) + len(self.text_feature_names) + len_time_columns
        )


def process_columns(dataset: pd.DataFrame, column_mapping: ColumnMapping) -> DatasetColumns:
    if column_mapping is None:
        # data mapping should not be empty in this step
        raise ValueError("column_mapping should be present")
    date_column = column_mapping.datetime if column_mapping.datetime in dataset else None
    # index column name
    id_column = column_mapping.id

    task = column_mapping.task

    target_column = column_mapping.target if column_mapping.target in dataset else None
    if task is None and target_column is not None:
        task = recognize_task(target_name=target_column, dataset=dataset)
    target_type = _get_target_type(dataset, column_mapping, task)
    num_feature_names = column_mapping.numerical_features
    cat_feature_names = column_mapping.categorical_features
    datetime_feature_names = column_mapping.datetime_features
    target_names = column_mapping.target_names
    utility_columns = [date_column, id_column, target_column]
    text_feature_names = column_mapping.text_features

    prediction_column: Optional[str] = None
    if isinstance(column_mapping.prediction, str):
        if column_mapping.prediction in dataset:
            prediction_column = column_mapping.prediction
        else:
            prediction_column = None
        utility_columns.append(prediction_column)
    elif column_mapping.prediction is None:
        prediction_column = None
    else:
        prediction_column = dataset[column_mapping.prediction].columns.tolist()

        if prediction_column:
            utility_columns += prediction_column

    utility_columns_set = set(utility_columns)
    cat_feature_names_set = set(cat_feature_names or [])
    text_feature_names_set = set(text_feature_names or [])

    if num_feature_names is None:
        # try to guess about numeric features in the dataset
        # ignore prediction, target, index and explicitly specified category columns and columns with text
        num_feature_names = sorted(
            list(
                set(dataset.select_dtypes([np.number]).columns)
                - utility_columns_set
                - cat_feature_names_set
                - text_feature_names_set
            )
        )

    else:
        empty_cols = dataset[num_feature_names].isnull().mean()
        empty_cols = empty_cols[empty_cols == 1.0].index
        num_feature_names = sorted(
            list(set(dataset[num_feature_names].select_dtypes([np.number]).columns).union(set(empty_cols)))
        )

    if datetime_feature_names is None:
        datetime_feature_names = sorted(list(set(dataset.select_dtypes(["datetime"]).columns) - utility_columns_set))
    else:
        empty_cols = dataset[datetime_feature_names].isnull().mean()
        empty_cols = empty_cols[empty_cols == 1.0].index
        datetime_feature_names = sorted(
            list(set(dataset[datetime_feature_names].select_dtypes(["datetime"]).columns).union(set(empty_cols)))
        )

    cat_feature_names = column_mapping.categorical_features

    if cat_feature_names is None:
        cat_feature_names = sorted(
            list(
                set(dataset.select_dtypes(exclude=[np.number, "datetime"]).columns)
                - utility_columns_set
                - text_feature_names_set
            )
        )

    else:
        cat_feature_names = dataset[cat_feature_names].columns.tolist()

    return DatasetColumns(
        DatasetUtilityColumns(date_column, id_column, target_column, prediction_column),
        target_type=target_type,
        num_feature_names=num_feature_names or [],
        cat_feature_names=cat_feature_names or [],
        datetime_feature_names=datetime_feature_names or [],
        target_names=target_names,
        task=task,
        text_feature_names=text_feature_names or [],
    )


def _get_target_type(dataset: pd.DataFrame, column_mapping: ColumnMapping, task: Optional[str]) -> Optional[str]:
    """
    Args:
        dataset: input dataset
        column_mapping: column definition from user
    Returns:
        type of target (or prediction, if there are only prediction) or None if both columns missing.
    """
    column = None
    if column_mapping.target is not None and column_mapping.target in dataset:
        column = dataset[column_mapping.target]
    if (
        column is None
        and column_mapping.prediction is not None
        and isinstance(column_mapping.prediction, str)
        and column_mapping.prediction in dataset
    ):
        column = dataset[column_mapping.prediction]

    if column is None:
        return None

    if column_mapping.target_names is not None or task == "classification":
        column_type = "cat"
    elif pd.api.types.is_numeric_dtype(column.dtype):
        column_type = "num"
    elif pd.api.types.is_datetime64_dtype(column.dtype):
        column_type = "datetime"
    else:
        column_type = "cat"
    return column_type


def recognize_task(target_name: str, dataset: pd.DataFrame) -> str:
    """Try to guess about the target type:
    if the target has a numeric type and number of unique values > 5: task == ‘regression’
    in all other cases task == ‘classification’.

    Args:
        target_name: name of target column.
        dataset: usually the data which you used in training.

    Returns:
        Task parameter.
    """
    if pd.api.types.is_numeric_dtype(dataset[target_name]) and dataset[target_name].nunique() >= 5:
        task = "regression"

    else:
        task = "classification"

    return task


def recognize_column_type(
    dataset: pd.DataFrame,
    column_name: str,
    columns: DatasetColumns,
) -> str:
    """Try to get the column type."""
    column = dataset[column_name]
    reg_condition = columns.task == "regression" or (
        pd.api.types.is_numeric_dtype(column) and columns.task != "classification" and column.nunique() > 5
    )
    if column_name == columns.utility_columns.target:
        if reg_condition:
            return "num"

        else:
            return "cat"

    if column_name == columns.utility_columns.prediction:

        if reg_condition:
            return "num"

        else:
            return "cat"

    if column_name in columns.num_feature_names:
        return "num"

    if isinstance(columns.utility_columns.prediction, list) and column_name in columns.utility_columns.prediction:
        return "num"

    if column_name in columns.cat_feature_names:
        return "cat"

    if column_name in columns.datetime_feature_names:
        return "datetime"

    if column_name in columns.text_feature_names:
        return "text"

    if column_name == columns.utility_columns.id_column:
        return "id"

    if column_name == columns.utility_columns.date:
        return "date"

    return "unknown"



import dataclasses
from enum import Enum
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union
from uuid import uuid4

import numpy as np
import pandas as pd
from plotly import graph_objs as go
from plotly.subplots import make_subplots

from evidently.model.widget import BaseWidgetInfo
from evidently.model.widget import PlotlyGraphInfo
from evidently.model.widget import TabInfo
from evidently.model.widget import WidgetType
from evidently.options import ColorOptions
from evidently.utils.types import Numeric
from evidently.utils.visualizations import Distribution


class WidgetSize(Enum):
    HALF = 1
    FULL = 2


class GraphData:
    title: str
    data: dict
    layout: dict

    def __init__(self, title: str, data: dict, layout: dict):
        """
        create GraphData object for usage in plotly_graph_tabs or plotly_data.

        Args:
            title: title of graph
            data: plotly figure data
            layout: plotly figure layout
        """
        self.title = title
        self.data = data
        self.layout = layout

    @staticmethod
    def figure(title: str, figure: go.Figure):
        """
        create GraphData from plotly figure itself
        Args:
            title: title of graph
            figure: plotly figure for getting data from
        """
        data = figure.to_plotly_json()
        return GraphData(title, data["data"], data["layout"])


def plotly_graph(*, graph_data: GraphData, size: WidgetSize = WidgetSize.FULL) -> BaseWidgetInfo:
    """
    generate plotly plot with given GraphData object.

    Args:
        graph_data: plot data for widget
        size: size of widget to render

    Example:
        >>> figure = go.Figure(go.Bar(name="Bar plot", x=[1, 2, 3, 4], y=[10, 11, 20, 11]))
        >>> f_dict = figure.to_plotly_json()
        >>> bar_graph_data = GraphData(title="Some plot title", data=f_dict["data"], layout=f_dict["layout"])
        >>> widget_info = plotly_graph(graph_data=bar_graph_data, size=WidgetSize.FULL)
    """
    return BaseWidgetInfo(
        title=graph_data.title,
        type=WidgetType.BIG_GRAPH.value,
        size=size.value,
        params={"data": graph_data.data, "layout": graph_data.layout},
    )


def plotly_data(*, title: str, data: dict, layout: dict, size: WidgetSize = WidgetSize.FULL) -> BaseWidgetInfo:
    """
    generate plotly plot with given data and layout (can be generated from plotly).

    Args:
        title: widget title
        data: plotly figure data
        layout: plotly figure layout
        size: widget size

    Example:
        >>> figure = go.Figure(go.Bar(name="Bar plot", x=[1, 2, 3, 4], y=[10, 11, 20, 11]))
        >>> f_dict = figure.to_plotly_json()
        >>> widget_info = plotly_data(title="Some plot title", data=f_dict["data"], layout=f_dict["layout"])
    """
    return plotly_graph(graph_data=GraphData(title, data, layout), size=size)


def plotly_figure(*, title: str, figure: go.Figure, size: WidgetSize = WidgetSize.FULL) -> BaseWidgetInfo:
    """
    generate plotly plot based on given plotly figure object.

    Args:
        title: title of widget
        figure: plotly figure which should be rendered as widget
        size: size of widget, default to WidgetSize.FULL

    Example:
        >>> bar_figure = go.Figure(go.Bar(name="Bar plot", x=[1, 2, 3, 4], y=[10, 11, 20, 11]))
        >>> widget_info = plotly_figure(title="Bar plot widget", figure=bar_figure, size=WidgetSize.FULL)
    """
    return plotly_graph(graph_data=GraphData.figure(title=title, figure=figure), size=size)


def plotly_graph_tabs(*, title: str, figures: List[GraphData], size: WidgetSize = WidgetSize.FULL) -> BaseWidgetInfo:
    """
    generate Tab widget with multiple graphs

    Args:
        title: widget title
        figures: list of graphs with tab titles
        size: widget size

    Example:
        >>> bar_figure = go.Figure(go.Bar(name="Bar plot", x=[1, 2, 3, 4], y=[10, 11, 20, 11]))
        >>> line_figure = go.Figure(go.Line(name="Bar plot", x=[1, 2, 3, 4], y=[10, 11, 20, 11]))
        >>> widget_info = plotly_graph_tabs(
        ...     title="Tabbed widget",
        ...     figures=[GraphData.figure("Bar", bar_figure), GraphData.figure("Line", line_figure)],
        ... )
    """
    return BaseWidgetInfo(
        title=title,
        type=WidgetType.TABBED_GRAPH.value,
        size=size.value,
        params={
            "graphs": [
                {
                    "id": str(uuid4()),
                    "title": graph.title,
                    "graph": {
                        "data": graph.data,
                        "layout": graph.layout,
                    },
                }
                for graph in figures
            ]
        },
    )


class CounterData:
    label: str
    value: str

    def __init__(self, label: str, value: str):
        """
        creates CounterData for counter widget with given label and value.

        Args:
            label: counter label
            value: counter value
        """
        self.label = label
        self.value = value

    @staticmethod
    def float(label: str, value: float, precision: int) -> "CounterData":
        """
        create CounterData for float value with given precision.

        Args:
            label: counter label
            value: float value of counter
            precision: decimal precision
        """
        return CounterData(label, f"{value:.{precision}}")

    @staticmethod
    def string(label: str, value: str) -> "CounterData":
        """
        create CounterData for string value with given precision.

        Args:
            label: counter label
            value: string value of counter
        """
        return CounterData(label, f"{value}")

    @staticmethod
    def int(label: str, value: int) -> "CounterData":
        """
        create CounterData for int value.

        Args:
            label: counter label
            value: int value
        """
        return CounterData(label, f"{value}")


def counter(
    *,
    counters: List[CounterData],
    title: str = "",
    size: WidgetSize = WidgetSize.FULL,
) -> BaseWidgetInfo:
    """
    generate widget with given counters

    Args:
        title: widget title
        counters: list of counters in widget
        size: widget size

    Example:
        >>> display_counters = [CounterData("value1", "some value"), CounterData.float("float", 0.111, 2)]
        >>> widget_info = counter(counters=display_counters, title="counters example")
    """
    return BaseWidgetInfo(
        title=title,
        type=WidgetType.COUNTER.value,
        size=size.value,
        params={"counters": [{"value": item.value, "label": item.label} for item in counters]},
    )


def header_text(*, label: str, title: str = "", size: WidgetSize = WidgetSize.FULL):
    """
    generate widget with some text as header

    Args:
        label: text to display
        title: widget title
        size: widget size
    """
    return BaseWidgetInfo(
        title=title,
        type=WidgetType.COUNTER.value,
        size=size.value,
        params={"counters": [{"value": "", "label": label}]},
    )


def table_data(
    *, column_names: Iterable[str], data: Iterable[Iterable], title: str = "", size: WidgetSize = WidgetSize.FULL
) -> BaseWidgetInfo:
    """
    generate simple table with given columns and data

    Args:
        column_names: list of column names in display order
        data: list of data rows (lists of object to show in table in order of columns), object will be converted to str
        title: widget title
        size: widget size

    Example:
        >>> columns = ["Column A", "Column B"]
        >>> in_table_data = [[1, 2], [3, 4]]
        >>> widget_info = table_data(column_names=columns, data=in_table_data, title="Table")
    """
    return BaseWidgetInfo(
        title=title,
        type=WidgetType.TABLE.value,
        params={
            "header": column_names,
            "data": [[str(item) for item in row] for row in data],
        },
        size=size.value,
    )


class ColumnType(Enum):
    STRING = "string"
    LINE = "line"
    SCATTER = "scatter"
    HISTOGRAM = "histogram"


class SortDirection(Enum):
    ASC = "asc"
    DESC = "desc"


@dataclasses.dataclass
class ColumnDefinition:
    title: str
    field_name: str
    type: ColumnType = ColumnType.STRING
    sort: Optional[SortDirection] = None
    options: Optional[dict] = None

    def as_dict(self) -> dict:
        result: dict = {"title": self.title, "field": self.field_name}
        if self.type != ColumnType.STRING:
            result["type"] = self.type.value
        if self.sort is not None:
            result["sort"] = self.sort.value
        if self.options is not None:
            result["options"] = self.options
        return result


@dataclasses.dataclass
class TabData:
    title: str
    widget: BaseWidgetInfo


def widget_tabs(*, title: str = "", size: WidgetSize = WidgetSize.FULL, tabs: List[TabData]) -> BaseWidgetInfo:
    """
    generate widget with tabs which can contain any other widget.

    Args:
        title: widget title
        size: widget size
        tabs: list of TabData with widgets to include

    Example:
        >>> columns = ["Column A", "Column B"]
        >>> in_table_data = [[1, 2], [3, 4]]
        >>> tab_data = [
        ...     TabData("Counters", counter(counters=[CounterData("counter", "value")], title="Counter")),
        ...     TabData("Table", table_data(column_names=columns, data=in_table_data, title="Table")),
        ... ]
        >>> widget_info = widget_tabs(title="Tabs", tabs=tab_data)
    """
    return BaseWidgetInfo(
        title=title,
        type=WidgetType.TABS.value,
        size=size.value,
        tabs=[TabInfo(str(uuid4()), tab.title, tab.widget) for tab in tabs],
    )


def widget_tabs_for_more_than_one(
    *, title: str = "", size: WidgetSize = WidgetSize.FULL, tabs: List[TabData]
) -> Optional[BaseWidgetInfo]:
    """Draw tabs widget only if there is more than one tab, otherwise just draw one widget"""
    if len(tabs) > 1:
        return widget_tabs(title=title, size=size, tabs=tabs)

    elif len(tabs) < 1:
        return None

    else:
        return tabs[0].widget


class DetailsPartInfo:
    title: str
    info: Union[BaseWidgetInfo, PlotlyGraphInfo]

    def __init__(self, title: str, info: Union[BaseWidgetInfo, PlotlyGraphInfo]):
        self.title = title
        self.info = info


class RowDetails:
    parts: List[DetailsPartInfo]

    def __init__(self, parts: Optional[List[DetailsPartInfo]] = None):
        if parts is None:
            parts = []
        self.parts = parts

    def with_part(self, title: str, info: Union[BaseWidgetInfo, PlotlyGraphInfo]):
        self.parts.append(DetailsPartInfo(title, info))
        return self


class RichTableDataRow:
    details: Optional[RowDetails]
    fields: dict

    def __init__(self, fields: dict, details: Optional[RowDetails] = None):
        self.fields = fields
        self.details = details


def rich_table_data(
    *,
    title: str = "",
    size: WidgetSize = WidgetSize.FULL,
    rows_per_page: int = 10,
    columns: List[ColumnDefinition],
    data: List[RichTableDataRow],
) -> BaseWidgetInfo:
    """
    generate widget with rich table: with additional column types and details for rows

    Args:
         title: widget title
         size: widget size
         rows_per_page: maximum number per page to show
         columns: list of columns in table
         data: list of dicts with data (key-value pairs, keys is according to ColumnDefinition.field_name)

    Example:
        >>> columns_def = [
        ...     ColumnDefinition("Column A", "field_1"),
        ...     ColumnDefinition("Column B", "field_2", ColumnType.HISTOGRAM,
        ...                      options={"xField": "x", "yField": "y", "color": "#ed0400"}),
        ...     ColumnDefinition("Column C", "field_3", sort=SortDirection.ASC),
        ... ]
        >>> in_table_data = [
        ...     RichTableDataRow(fields=dict(field_1="a", field_2=dict(x=[1, 2, 3], y=[10, 11, 3]), field_3="2")),
        ...     RichTableDataRow(
        ...         fields=dict(field_1="b", field_2=dict(x=[1, 2, 3], y=[10, 11, 3]), field_3="1"),
        ...         details=RowDetails()
        ...             .with_part("Some details", counter(counters=[CounterData("counter 1", "value")])
        ...         )
        ...     )
        ... ]
        >>> widget_info = rich_table_data(title="Rich table", rows_per_page=10, columns=columns_def, data=in_table_data)
    """
    additional_graphs = []

    converted_data = []
    for row in data:
        if row.details is None or row.details.parts is None or len(row.details.parts) == 0:
            converted_data.append(dict(**row.fields))
            continue
        parts = []
        for part in row.details.parts:
            parts.append(
                dict(
                    title=part.title,
                    id=part.info.id,
                    type="widget" if isinstance(part.info, BaseWidgetInfo) else "graph",
                )
            )
            additional_graphs.append(part.info)
        converted_data.append(dict(details={"parts": parts}, **row.fields))

    return BaseWidgetInfo(
        title=title,
        type=WidgetType.BIG_TABLE.value,
        details="",
        alerts=[],
        alertsPosition="row",
        insights=[],
        size=size.value,
        params={
            "rowsPerPage": min(len(data), rows_per_page),
            "columns": [column.as_dict() for column in columns],
            "data": converted_data,
        },
        additionalGraphs=additional_graphs,
    )


@dataclasses.dataclass
class HistogramData:
    name: str
    x: list
    y: List[Union[int, float]]


def get_histogram_figure(
    *,
    primary_hist: HistogramData,
    secondary_hist: Optional[HistogramData] = None,
    color_options: ColorOptions,
    orientation: str = "v",
) -> go.Figure:
    figure = go.Figure()
    curr_bar = go.Bar(
        name=primary_hist.name,
        x=primary_hist.x,
        y=primary_hist.y,
        marker_color=color_options.get_current_data_color(),
        orientation=orientation,
    )
    figure.add_trace(curr_bar)

    if secondary_hist is not None:
        ref_bar = go.Bar(
            name=secondary_hist.name,
            x=secondary_hist.x,
            y=secondary_hist.y,
            marker_color=color_options.get_reference_data_color(),
            orientation=orientation,
        )
        figure.add_trace(ref_bar)

    return figure


def get_histogram_figure_with_range(
    *,
    primary_hist: HistogramData,
    secondary_hist: Optional[HistogramData] = None,
    left: Numeric,
    right: Numeric,
    orientation: str = "v",
    color_options: ColorOptions,
) -> go.Figure:
    figure = get_histogram_figure(
        primary_hist=primary_hist,
        secondary_hist=secondary_hist,
        color_options=color_options,
        orientation=orientation,
    )
    max_y = np.max([np.max(x["y"]) for x in figure.data])
    min_y = np.min([np.min(x["y"]) for x in figure.data])

    figure.add_trace(
        go.Scatter(
            x=[left, left],
            y=[min_y, max_y],
            mode="lines",
            line={"color": color_options.vertical_lines, "width": 2, "dash": "dash"},
            name="range left value",
        )
    )
    figure.add_trace(
        go.Scatter(
            x=[right, right],
            y=[min_y, max_y],
            mode="lines",
            line={"color": color_options.vertical_lines, "width": 2, "dash": "solid"},
            name="range right value",
        )
    )
    figure.add_vrect(x0=left, x1=right, fillcolor=color_options.fill_color, opacity=0.25, line_width=0)
    return figure


def get_histogram_figure_with_quantile(
    *,
    current: HistogramData,
    reference: Optional[HistogramData] = None,
    current_quantile: float,
    reference_quantile: Optional[float] = None,
    color_options: ColorOptions,
    orientation: str = "v",
) -> go.Figure:
    figure = get_histogram_figure(
        primary_hist=current,
        secondary_hist=reference,
        color_options=color_options,
        orientation=orientation,
    )
    # add quantile lines. Use scatter, not `add_vline`
    # because `add_vline` doesn't support legend, it is not interactive
    max_y = np.max([np.max(x["y"]) for x in figure.data])
    min_y = np.min([np.min(x["y"]) for x in figure.data])

    figure.add_trace(
        go.Scatter(
            x=[current_quantile, current_quantile],
            y=[min_y, max_y],
            mode="lines",
            line={"color": color_options.vertical_lines, "width": 2, "dash": "dash"},
            name="reference quantile",
        )
    )

    if reference_quantile is not None:
        figure.add_trace(
            go.Scatter(
                x=[reference_quantile, reference_quantile],
                y=[min_y, max_y],
                mode="lines",
                line={"color": color_options.vertical_lines, "width": 2, "dash": "solid"},
                name="current quantile",
            )
        )

    return figure


def histogram(
    *,
    title: str,
    primary_hist: HistogramData,
    secondary_hist: Optional[HistogramData] = None,
    color_options: ColorOptions,
    orientation: str = "v",
    size: WidgetSize = WidgetSize.FULL,
    xaxis_title: Optional[str] = None,
    yaxis_title: Optional[str] = None,
) -> BaseWidgetInfo:
    """
    generate widget with one or two histogram

    Args:
        title: widget title
        primary_hist: first histogram to show in widget
        secondary_hist: optional second histogram to show in widget
        orientation: bars orientation in histograms
        color_options: color options to use for widgets
        size: widget size
        xaxis_title: title for x-axis
        yaxis_title: title for y-axis
    Example:
        >>> ref_hist = HistogramData("Histogram 1", x=["a", "b", "c"], y=[1, 2, 3])
        >>> curr_hist = HistogramData("Histogram 2", x=["a", "b", "c"], y=[3, 2 ,1])
        >>> widget_info = histogram(
        >>>     title="Histogram example",
        >>>     primary_hist=ref_hist,
        >>>     secondary_hist=curr_hist,
        >>>     color_options=color_options
        >>> )
    """
    figure = get_histogram_figure(
        primary_hist=primary_hist,
        secondary_hist=secondary_hist,
        color_options=color_options,
        orientation=orientation,
    )
    if xaxis_title is not None:
        figure.update_layout(
            xaxis_title=xaxis_title,
        )

    if yaxis_title is not None:
        figure.update_layout(
            yaxis_title=yaxis_title,
        )

    return plotly_figure(title=title, figure=figure, size=size)


def get_histogram_for_distribution(
    *,
    current_distribution: Distribution,
    reference_distribution: Optional[Distribution] = None,
    title: str = "",
    xaxis_title: Optional[str] = None,
    yaxis_title: Optional[str] = None,
    color_options: ColorOptions,
):
    current_histogram = HistogramData(
        name="current",
        x=current_distribution.x,
        y=current_distribution.y,
    )

    if reference_distribution is not None:
        reference_histogram: Optional[HistogramData] = HistogramData(
            name="reference",
            x=reference_distribution.x,
            y=reference_distribution.y,
        )

    else:
        reference_histogram = None

    return histogram(
        title=title,
        primary_hist=current_histogram,
        secondary_hist=reference_histogram,
        xaxis_title=xaxis_title,
        yaxis_title=yaxis_title,
        color_options=color_options,
    )


@dataclasses.dataclass
class HeatmapData:
    name: str
    matrix: pd.DataFrame


def get_heatmaps_widget(
    *,
    title: str = "",
    primary_data: HeatmapData,
    secondary_data: Optional[HeatmapData] = None,
    size: WidgetSize = WidgetSize.FULL,
    color_options: ColorOptions,
) -> BaseWidgetInfo:
    """
    Create a widget with heatmap(s)
    """

    if secondary_data is not None:
        subplot_titles = [primary_data.name, secondary_data.name]
        heatmaps_count = 2

    else:
        subplot_titles = [""]
        heatmaps_count = 1

    figure = make_subplots(rows=1, cols=heatmaps_count, subplot_titles=subplot_titles, shared_yaxes=True)

    for idx, heatmap_data in enumerate([primary_data, secondary_data]):
        if heatmap_data is None:
            continue
        data = heatmap_data.matrix
        columns = heatmap_data.matrix.columns

        # show values if thw heatmap is small
        if len(columns) < 15:
            heatmap_text = np.round(data, 2).astype(str)
            heatmap_text_template: Optional[str] = "%{text}"

        else:
            heatmap_text = None
            heatmap_text_template = None

        figure.add_trace(
            go.Heatmap(
                z=data,
                x=columns,
                y=columns,
                text=heatmap_text,
                texttemplate=heatmap_text_template,
                coloraxis="coloraxis",
            ),
            1,
            idx + 1,
        )

    figure.update_layout(coloraxis={"colorscale": color_options.heatmap})
    figure.update_yaxes(type="category")
    figure.update_xaxes(tickangle=-45)
    return plotly_figure(title=title, figure=figure, size=size)


def get_roc_auc_tab_data(
    curr_roc_curve: dict, ref_roc_curve: Optional[dict], color_options: ColorOptions
) -> List[Tuple[str, BaseWidgetInfo]]:
    additional_plots = []
    cols = 1
    subplot_titles = [""]
    if ref_roc_curve is not None:
        cols = 2
        subplot_titles = ["current", "reference"]
    for label in curr_roc_curve.keys():
        fig = make_subplots(rows=1, cols=cols, subplot_titles=subplot_titles, shared_yaxes=True)
        trace = go.Scatter(
            x=curr_roc_curve[label]["fpr"],
            y=curr_roc_curve[label]["tpr"],
            mode="lines",
            name="ROC",
            legendgroup="ROC",
            marker=dict(
                size=6,
                color=color_options.get_current_data_color(),
            ),
        )
        fig.add_trace(trace, 1, 1)
        fig.update_xaxes(title_text="False Positive Rate", row=1, col=1)
        if ref_roc_curve is not None:
            trace = go.Scatter(
                x=ref_roc_curve[label]["fpr"],
                y=ref_roc_curve[label]["tpr"],
                mode="lines",
                name="ROC",
                legendgroup="ROC",
                showlegend=False,
                marker=dict(
                    size=6,
                    color=color_options.get_current_data_color(),
                ),
            )
            fig.add_trace(trace, 1, 2)
            fig.update_xaxes(title_text="False Positive Rate", row=1, col=2)
        fig.update_layout(yaxis_title="True Positive Rate", showlegend=True)

        additional_plots.append((str(label), plotly_figure(title="", figure=fig)))
    return additional_plots


def get_pr_rec_plot_data(
    current_pr_curve: dict, reference_pr_curve: Optional[dict], color_options: ColorOptions
) -> List[Tuple[str, BaseWidgetInfo]]:
    additional_plots = []
    cols = 1
    subplot_titles = [""]
    if reference_pr_curve is not None:
        cols = 2
        subplot_titles = ["current", "reference"]
    for label in current_pr_curve.keys():
        fig = make_subplots(rows=1, cols=cols, subplot_titles=subplot_titles, shared_yaxes=True)
        trace = go.Scatter(
            x=current_pr_curve[label]["rcl"],
            y=current_pr_curve[label]["pr"],
            mode="lines",
            name="PR",
            legendgroup="PR",
            marker=dict(
                size=6,
                color=color_options.get_current_data_color(),
            ),
        )
        fig.add_trace(trace, 1, 1)
        fig.update_xaxes(title_text="Recall", row=1, col=1)
        if reference_pr_curve is not None:
            trace = go.Scatter(
                x=reference_pr_curve[label]["rcl"],
                y=reference_pr_curve[label]["pr"],
                mode="lines",
                name="PR",
                legendgroup="PR",
                showlegend=False,
                marker=dict(
                    size=6,
                    color=color_options.get_current_data_color(),
                ),
            )
            fig.add_trace(trace, 1, 2)
            fig.update_xaxes(title_text="Recall", row=1, col=2)
        fig.update_layout(yaxis_title="Precision", showlegend=True)

        additional_plots.append((str(label), plotly_figure(title="", figure=fig)))
    return additional_plots


def get_class_separation_plot_data(
    current_plot: pd.DataFrame, reference_plot: Optional[pd.DataFrame], target_name: str, color_options: ColorOptions
) -> List[Tuple[str, BaseWidgetInfo]]:
    additional_plots = []
    cols = 1
    subplot_titles = [""]
    if reference_plot is not None:
        cols = 2
        subplot_titles = ["current", "reference"]
    for label in current_plot.columns.drop(target_name):
        fig = make_subplots(rows=1, cols=cols, subplot_titles=subplot_titles, shared_yaxes=True)
        trace = go.Scatter(
            x=np.random.random(current_plot[current_plot[target_name] == label].shape[0]),
            y=current_plot[current_plot[target_name] == label][label],
            mode="markers",
            name=str(label),
            legendgroup=str(label),
            marker=dict(size=6, color=color_options.primary_color),
        )
        fig.add_trace(trace, 1, 1)

        trace = go.Scatter(
            x=np.random.random(current_plot[current_plot[target_name] != label].shape[0]),
            y=current_plot[current_plot[target_name] != label][label],
            mode="markers",
            name="other",
            legendgroup="other",
            marker=dict(size=6, color=color_options.secondary_color),
        )
        fig.add_trace(trace, 1, 1)
        fig.update_xaxes(dict(range=(-2, 3), showticklabels=False), row=1, col=1)

        if reference_plot is not None:
            trace = go.Scatter(
                x=np.random.random(reference_plot[reference_plot[target_name] == label].shape[0]),
                y=reference_plot[reference_plot[target_name] == label][label],
                mode="markers",
                name=str(label),
                legendgroup=str(label),
                showlegend=False,
                marker=dict(size=6, color=color_options.primary_color),
            )
            fig.add_trace(trace, 1, 2)

            trace = go.Scatter(
                x=np.random.random(reference_plot[reference_plot[target_name] != label].shape[0]),
                y=reference_plot[reference_plot[target_name] != label][label],
                mode="markers",
                name="other",
                legendgroup="other",
                showlegend=False,
                marker=dict(size=6, color=color_options.secondary_color),
            )
            fig.add_trace(trace, 1, 2)
            fig.update_xaxes(dict(range=(-2, 3), showticklabels=False), row=1, col=2)
        fig.update_layout(yaxis_title="Probability", showlegend=True)

        additional_plots.append((str(label), plotly_figure(title="", figure=fig)))
    return additional_plots



from dataclasses import dataclass
from typing import Dict

import numpy as np
import pandas as pd
from scipy.stats import probplot

from evidently.utils.data_operations import DatasetColumns


class ErrorWithQuantiles:
    def __init__(self, error, quantile_top, quantile_other):
        self.error = error
        self.quantile_top = quantile_top
        self.quantile_other = quantile_other


@dataclass
class FeatureBias:
    feature_type: str
    majority: float
    under: float
    over: float
    range: float

    def as_dict(self, prefix):
        return {
            prefix + "majority": self.majority,
            prefix + "under": self.under,
            prefix + "over": self.over,
            prefix + "range": self.range,
        }


def _calculate_error_normality(error: ErrorWithQuantiles):
    qq_lines = probplot(error.error, dist="norm", plot=None)
    qq_dots = [t.tolist() for t in qq_lines[0]]
    qq_line = list(qq_lines[1])
    return {
        "order_statistic_medians_x": [float(x) for x in qq_dots[0]],
        "order_statistic_medians_y": [float(x) for x in qq_dots[1]],
        "slope": float(qq_line[0]),
        "intercept": float(qq_line[1]),
        "r": float(qq_line[2]),
    }


def _calculate_quality_metrics(dataset, prediction_column, target_column, conf_interval_n_sigmas=1):
    me = np.mean(dataset[prediction_column] - dataset[target_column])
    sde = np.std(dataset[prediction_column] - dataset[target_column], ddof=1)

    abs_err = np.abs(dataset[prediction_column] - dataset[target_column])
    abs_error_max = abs_err.max()
    mae = np.mean(abs_err)
    sdae = np.std(abs_err, ddof=1)

    epsilon = np.finfo(np.float64).eps
    abs_perc_err = np.abs(dataset[prediction_column] - dataset[target_column]) / np.maximum(
        dataset[target_column], epsilon
    )
    mape = 100.0 * np.mean(abs_perc_err)
    sdape = np.std(abs_perc_err, ddof=1)

    return {
        "mean_error": float(me),
        "mean_abs_error": float(mae),
        "mean_abs_perc_error": float(mape),
        "abs_error_max": abs_error_max,
        "error_std": conf_interval_n_sigmas * float(sde),
        "abs_error_std": conf_interval_n_sigmas * float(sdae),
        "abs_perc_error_std": conf_interval_n_sigmas * float(sdape),
    }


def _prepare_dataset(dataset, target_column, prediction_column):
    dataset.replace([np.inf, -np.inf], np.nan, inplace=True)
    dataset.dropna(axis=0, how="any", inplace=True, subset=[target_column, prediction_column])


def _calculate_underperformance(err_quantiles: ErrorWithQuantiles, conf_interval_n_sigmas: int = 1):
    error = err_quantiles.error
    quantile_top = err_quantiles.quantile_top
    quantile_other = err_quantiles.quantile_other
    mae_under = np.mean(error[error <= quantile_top])
    mae_exp = np.mean(error[(error > quantile_top) & (error < quantile_other)])
    mae_over = np.mean(error[error >= quantile_other])
    sd_under = np.std(error[error <= quantile_top], ddof=1)
    sd_exp = np.std(error[(error > quantile_top) & (error < quantile_other)], ddof=1)
    sd_over = np.std(error[error >= quantile_other], ddof=1)

    return {
        "majority": {"mean_error": float(mae_exp), "std_error": conf_interval_n_sigmas * float(sd_exp)},
        "underestimation": {"mean_error": float(mae_under), "std_error": conf_interval_n_sigmas * float(sd_under)},
        "overestimation": {"mean_error": float(mae_over), "std_error": conf_interval_n_sigmas * float(sd_over)},
    }


def error_bias_table(dataset, err_quantiles, num_feature_names, cat_feature_names) -> Dict[str, FeatureBias]:
    num_bias = {
        feature_name: _error_num_feature_bias(dataset, feature_name, err_quantiles)
        for feature_name in num_feature_names
    }
    cat_bias = {
        feature_name: _error_cat_feature_bias(dataset, feature_name, err_quantiles)
        for feature_name in cat_feature_names
    }
    error_bias = num_bias.copy()
    error_bias.update(cat_bias)
    return error_bias


def _error_num_feature_bias(dataset, feature_name, err_quantiles: ErrorWithQuantiles) -> FeatureBias:
    error = err_quantiles.error
    quantile_top = err_quantiles.quantile_top
    quantile_other = err_quantiles.quantile_other
    ref_overal_value = np.mean(dataset[feature_name])
    ref_under_value = np.mean(dataset[error <= quantile_top][feature_name])

    ref_over_value = np.mean(dataset[error >= quantile_other][feature_name])
    if ref_over_value == ref_under_value:
        ref_range_value = 0

    else:
        ref_range_value = (
            100
            * abs(ref_over_value - ref_under_value)
            / (np.max(dataset[feature_name]) - np.min(dataset[feature_name]))
        )

    return FeatureBias(
        feature_type="num",
        majority=float(ref_overal_value),
        under=float(ref_under_value),
        over=float(ref_over_value),
        range=float(ref_range_value),
    )


def _stable_value_counts(series: pd.Series):
    return series.value_counts().reindex(pd.unique(series))


def _error_cat_feature_bias(dataset, feature_name, err_quantiles: ErrorWithQuantiles) -> FeatureBias:
    error = err_quantiles.error
    quantile_top = err_quantiles.quantile_top
    quantile_other = err_quantiles.quantile_other
    ref_overall_value = _stable_value_counts(dataset[feature_name]).idxmax()
    ref_under_value = _stable_value_counts(dataset[error <= quantile_top][feature_name]).idxmax()
    ref_over_value = _stable_value_counts(dataset[error >= quantile_other][feature_name]).idxmax()
    if (
        (ref_overall_value != ref_under_value)
        or (ref_over_value != ref_overall_value)
        or (ref_under_value != ref_overall_value)
    ):
        ref_range_value = 1
    else:
        ref_range_value = 0

    if pd.isnull(ref_overall_value):
        ref_overall_value = None

    if pd.isnull(ref_under_value):
        ref_under_value = None

    if pd.isnull(ref_over_value):
        ref_over_value = None

    return FeatureBias(
        feature_type="cat",
        majority=ref_overall_value,
        under=ref_under_value,
        over=ref_over_value,
        range=float(ref_range_value),
    )


def error_with_quantiles(dataset, prediction_column, target_column, quantile: float):
    error = dataset[prediction_column] - dataset[target_column]

    # underperformance metrics
    quantile_top = np.quantile(error, quantile)
    quantile_other = np.quantile(error, 1 - quantile)
    return ErrorWithQuantiles(error, quantile_top, quantile_other)


@dataclass
class RegressionPerformanceMetrics:
    mean_error: float
    mean_abs_error: float
    mean_abs_perc_error: float
    error_std: float
    abs_error_max: float
    abs_error_std: float
    abs_perc_error_std: float
    error_normality: dict
    underperformance: dict
    error_bias: dict


def calculate_regression_performance(
    dataset: pd.DataFrame, columns: DatasetColumns, error_bias_prefix: str
) -> RegressionPerformanceMetrics:
    target_column = columns.utility_columns.target
    prediction_column = columns.utility_columns.prediction

    num_feature_names = columns.num_feature_names
    cat_feature_names = columns.cat_feature_names

    if target_column is None or prediction_column is None:
        raise ValueError("Target and prediction should be present")

    _prepare_dataset(dataset, target_column, prediction_column)
    # calculate quality metrics
    quality_metrics = _calculate_quality_metrics(dataset, prediction_column, target_column)
    # error normality
    err_quantiles = error_with_quantiles(dataset, prediction_column, target_column, quantile=0.05)
    quality_metrics["error_normality"] = _calculate_error_normality(err_quantiles)
    # underperformance metrics
    quality_metrics["underperformance"] = _calculate_underperformance(err_quantiles)
    quality_metrics["error_bias"] = {}
    feature_bias = error_bias_table(dataset, err_quantiles, num_feature_names, cat_feature_names)
    # convert to old format
    quality_metrics["error_bias"] = {
        feature: dict(feature_type=bias.feature_type, **bias.as_dict(error_bias_prefix))
        for feature, bias in feature_bias.items()
    }
    return RegressionPerformanceMetrics(**quality_metrics)



"""Available tests for TestSuite reports.
Tests grouped into modules.
For detailed information see module documentation.
"""

from .classification_performance_tests import TestAccuracyScore
from .classification_performance_tests import TestF1ByClass
from .classification_performance_tests import TestF1Score
from .classification_performance_tests import TestFNR
from .classification_performance_tests import TestFPR
from .classification_performance_tests import TestLogLoss
from .classification_performance_tests import TestPrecisionByClass
from .classification_performance_tests import TestPrecisionScore
from .classification_performance_tests import TestRecallByClass
from .classification_performance_tests import TestRecallScore
from .classification_performance_tests import TestRocAuc
from .classification_performance_tests import TestTNR
from .classification_performance_tests import TestTPR
from .data_drift_tests import TestAllFeaturesValueDrift
from .data_drift_tests import TestColumnDrift
from .data_drift_tests import TestCustomFeaturesValueDrift
from .data_drift_tests import TestNumberOfDriftedColumns
from .data_drift_tests import TestShareOfDriftedColumns
from .data_integrity_tests import TestAllColumnsShareOfMissingValues
from .data_integrity_tests import TestColumnAllConstantValues
from .data_integrity_tests import TestColumnAllUniqueValues
from .data_integrity_tests import TestColumnNumberOfDifferentMissingValues
from .data_integrity_tests import TestColumnNumberOfMissingValues
from .data_integrity_tests import TestColumnRegExp
from .data_integrity_tests import TestColumnShareOfMissingValues
from .data_integrity_tests import TestColumnsType
from .data_integrity_tests import TestNumberOfColumns
from .data_integrity_tests import TestNumberOfColumnsWithMissingValues
from .data_integrity_tests import TestNumberOfConstantColumns
from .data_integrity_tests import TestNumberOfDifferentMissingValues
from .data_integrity_tests import TestNumberOfDuplicatedColumns
from .data_integrity_tests import TestNumberOfDuplicatedRows
from .data_integrity_tests import TestNumberOfEmptyColumns
from .data_integrity_tests import TestNumberOfEmptyRows
from .data_integrity_tests import TestNumberOfMissingValues
from .data_integrity_tests import TestNumberOfRows
from .data_integrity_tests import TestNumberOfRowsWithMissingValues
from .data_integrity_tests import TestShareOfColumnsWithMissingValues
from .data_integrity_tests import TestShareOfMissingValues
from .data_integrity_tests import TestShareOfRowsWithMissingValues
from .data_quality_tests import TestAllColumnsMostCommonValueShare
from .data_quality_tests import TestCatColumnsOutOfListValues
from .data_quality_tests import TestColumnQuantile
from .data_quality_tests import TestColumnValueMax
from .data_quality_tests import TestColumnValueMean
from .data_quality_tests import TestColumnValueMedian
from .data_quality_tests import TestColumnValueMin
from .data_quality_tests import TestColumnValueStd
from .data_quality_tests import TestConflictPrediction
from .data_quality_tests import TestConflictTarget
from .data_quality_tests import TestCorrelationChanges
from .data_quality_tests import TestHighlyCorrelatedColumns
from .data_quality_tests import TestMeanInNSigmas
from .data_quality_tests import TestMostCommonValueShare
from .data_quality_tests import TestNumberOfOutListValues
from .data_quality_tests import TestNumberOfOutRangeValues
from .data_quality_tests import TestNumberOfUniqueValues
from .data_quality_tests import TestNumColumnsMeanInNSigmas
from .data_quality_tests import TestNumColumnsOutOfRangeValues
from .data_quality_tests import TestPredictionFeaturesCorrelations
from .data_quality_tests import TestShareOfOutListValues
from .data_quality_tests import TestShareOfOutRangeValues
from .data_quality_tests import TestTargetFeaturesCorrelations
from .data_quality_tests import TestTargetPredictionCorrelation
from .data_quality_tests import TestUniqueValuesShare
from .data_quality_tests import TestValueList
from .data_quality_tests import TestValueRange
from .regression_performance_tests import TestValueAbsMaxError
from .regression_performance_tests import TestValueMAE
from .regression_performance_tests import TestValueMAPE
from .regression_performance_tests import TestValueMeanError
from .regression_performance_tests import TestValueR2Score
from .regression_performance_tests import TestValueRMSE



import warnings
from dataclasses import dataclass
from typing import Dict
from typing import Optional
from typing import Union

from evidently.calculations.stattests import PossibleStatTestType
from evidently.calculations.stattests import StatTest
from evidently.utils.data_drift_utils import resolve_stattest_threshold

DEFAULT_NBINSX = 10


@dataclass
class DataDriftOptions:
    """Configuration for Data Drift calculations.

    Args:
        confidence: Defines the confidence level for statistical tests.
                    Applies to all features (if passed as float) or certain features (if passed as dictionary).
                    (Deprecated) Use `threshold` to define confidence level for statistical
                    tests as more universal solution.
        threshold: Defines thresholds for statistical tests.
                   Applies to all features (if passed as float) or certain features (if passed as dictionary).
        drift_share: Sets the share of drifting features as a condition for Dataset Drift in the Data Drift report.
        nbinsx: Defines the number of bins in a histogram.
                Applies to all features (if passed as int) or certain features (if passed as dictionary).
        xbins: Defines the boundaries for the size of a specific bin in a histogram.
        feature_stattest_func: Defines a custom statistical test for drift detection in the Data Drift report.
                               Applies to all features (if passed as a function) or individual features (if a dict).
                               (Deprecated) Use `all_features_stattest` or `per_feature_stattest`.
        all_features_stattest: Defines a custom statistical test for drift detection in the Data Drift report
                               for all features.
        cat_features_stattest: Defines a custom statistical test for drift detection in the Data Drift report
                               for categorical features only.
        num_features_stattest: Defines a custom statistical test for drift detection in the Data Drift report
                               for numerical features only.
        per_feature_stattest: Defines a custom statistical test for drift detection in the Data Drift report
                              per feature.
        cat_target_stattest_func: Defines a custom statistical test to detect target drift in category target.
        num_target_stattest_func: Defines a custom statistical test to detect target drift in numeric target.
    """

    confidence: Optional[Union[float, Dict[str, float]]] = None
    threshold: Optional[Union[float, Dict[str, float]]] = None
    drift_share: float = 0.5
    nbinsx: Union[int, Dict[str, int]] = DEFAULT_NBINSX
    xbins: Optional[Dict[str, int]] = None

    feature_stattest_func: Optional[Union[PossibleStatTestType, Dict[str, PossibleStatTestType]]] = None

    all_features_stattest: Optional[PossibleStatTestType] = None
    cat_features_stattest: Optional[PossibleStatTestType] = None
    num_features_stattest: Optional[PossibleStatTestType] = None
    text_features_stattest: Optional[PossibleStatTestType] = None
    per_feature_stattest: Optional[Dict[str, PossibleStatTestType]] = None

    all_features_threshold: Optional[float] = None
    cat_features_threshold: Optional[float] = None
    num_features_threshold: Optional[float] = None
    text_features_threshold: Optional[float] = None
    per_feature_threshold: Optional[Dict[str, float]] = None

    cat_target_threshold: Optional[float] = None
    num_target_threshold: Optional[float] = None

    cat_target_stattest_func: Optional[PossibleStatTestType] = None
    num_target_stattest_func: Optional[PossibleStatTestType] = None

    def as_dict(self):
        return {
            "confidence": self.confidence,
            "drift_share": self.drift_share,
            "nbinsx": self.nbinsx,
            "xbins": self.xbins,
        }

    def _calculate_threshold(self, feature_name: str, feature_type: str) -> Optional[float]:
        if self.threshold is not None:
            if isinstance(self.threshold, float):
                return self.threshold

            if isinstance(self.threshold, dict):
                return self.threshold.get(feature_name)

            raise ValueError(f"DataDriftOptions.threshold is incorrect type {type(self.threshold)}")

        _, threshold = resolve_stattest_threshold(
            feature_name,
            feature_type,
            self.all_features_stattest,
            self.cat_features_stattest,
            self.num_features_stattest,
            self.text_features_stattest,
            self.per_feature_stattest,
            self.all_features_threshold,
            self.cat_features_threshold,
            self.num_features_threshold,
            self.text_features_threshold,
            self.per_feature_threshold,
        )
        return threshold

    def get_threshold(self, feature_name: str, feature_type: str) -> Optional[float]:
        threshold = self._calculate_threshold(feature_name, feature_type)

        if self.confidence is not None and threshold is not None:
            raise ValueError("Only DataDriftOptions.confidence or DataDriftOptions.threshold can be set")

        if self.confidence is not None:
            warnings.warn("DataDriftOptions.confidence is deprecated, use DataDriftOptions.threshold instead.")

            if isinstance(self.confidence, float):
                return 1.0 - self.confidence

            if isinstance(self.confidence, dict):
                override = self.confidence.get(feature_name)
                return None if override is None else 1.0 - override

            raise ValueError(f"DataDriftOptions.confidence is incorrect type {type(self.confidence)}")

        return threshold

    def get_nbinsx(self, feature_name: str) -> int:
        if isinstance(self.nbinsx, int):
            return self.nbinsx
        if isinstance(self.nbinsx, dict):
            return self.nbinsx.get(feature_name, DEFAULT_NBINSX)
        raise ValueError(f"DataDriftOptions.nbinsx is incorrect type {type(self.nbinsx)}")

    def get_feature_stattest_func(self, feature_name: str, feature_type: str) -> Optional[PossibleStatTestType]:
        if self.feature_stattest_func is not None and any(
            [
                self.all_features_stattest,
                self.cat_features_stattest,
                self.num_features_stattest,
                self.text_features_stattest,
                self.per_feature_stattest,
            ]
        ):
            raise ValueError(
                "Cannot use DataDriftOptions.feature_stattest_func along with any "
                "of DataDriftOptions.cat_stattest_func,"
                " DataDriftOptions.num_stattest_func,"
                " DataDriftOptions.text_stattest_func,"
                " DataDriftOptions.per_feature_stattest_func."
            )
        if self.feature_stattest_func is not None:
            warnings.warn(
                "DataDriftOptions.feature_stattest_func is deprecated use DataDriftOptions.stattest_func"
                " or DataDriftOptions.per_feature_stattest_func."
            )
            if callable(self.feature_stattest_func) or isinstance(self.feature_stattest_func, (StatTest, str)):
                return self.feature_stattest_func
            if isinstance(self.feature_stattest_func, dict):
                return self.feature_stattest_func.get(feature_name)
            return None
        stattest, _ = resolve_stattest_threshold(
            feature_name,
            feature_type,
            self.all_features_stattest,
            self.cat_features_stattest,
            self.num_features_stattest,
            self.text_features_stattest,
            self.per_feature_stattest,
            self.all_features_threshold,
            self.cat_features_threshold,
            self.num_features_threshold,
            self.text_features_threshold,
            self.per_feature_threshold,
        )
        return stattest

    def __hash__(self) -> int:
        """Calculate hash for data drift options - for using in metrics deduplication via dicts."""
        return str(self.as_dict()).__hash__()



#!/usr/bin/env python
# coding: utf-8
import json
from typing import List
from typing import Optional
from typing import Tuple

import numpy as np
import pandas as pd
import plotly.graph_objects as go
from plotly.subplots import make_subplots

from evidently import ColumnMapping
from evidently.analyzers.data_quality_analyzer import DataQualityAnalyzer
from evidently.analyzers.data_quality_analyzer import DataQualityAnalyzerResults
from evidently.calculations.data_quality import FeatureQualityStats
from evidently.dashboard.widgets.widget import Widget
from evidently.model.widget import AdditionalGraphInfo
from evidently.model.widget import BaseWidgetInfo
from evidently.options import ColorOptions


class DataQualityFeaturesWidget(Widget):
    period_prefix: str

    def analyzers(self):
        return [DataQualityAnalyzer]

    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results,
    ) -> Optional[BaseWidgetInfo]:
        color_options = self.options_provider.get(ColorOptions)
        self.period_prefix = ""
        is_current_data = current_data is not None
        data_quality_results = DataQualityAnalyzer.get_results(analyzers_results)
        columns = data_quality_results.columns.utility_columns
        target_column = columns.target
        target_type: Optional[str]

        if target_column:
            target_type = data_quality_results.reference_features_stats[target_column].feature_type

        else:
            target_type = None

        cat_feature_names = data_quality_results.columns.cat_feature_names
        date_column = columns.date
        self._transform_cat_features(reference_data, current_data, cat_feature_names, target_column, target_type)

        all_features = data_quality_results.columns.get_all_features_list(
            cat_before_num=True, include_datetime_feature=True
        )
        if date_column:
            all_features = [date_column] + all_features
        if target_column:
            all_features = [target_column] + all_features

        if is_current_data:
            metrics_values_headers = ["reference", "current"]

        else:
            metrics_values_headers = [""]

        widgets_list = []
        for feature_name in all_features:
            additional_graphs = []
            feature_type = data_quality_results.reference_features_stats[feature_name].feature_type
            fig_main_distr = self._plot_main_distr_figure(
                reference_data, current_data, feature_name, feature_type, color_options
            )
            parts = self.assemble_parts(target_column, date_column, feature_name, feature_type)
            # additional_graphs = []
            if date_column and feature_type != "datetime":
                freq = self._choose_agg_period(date_column, reference_data, current_data)
                reference_data[date_column + "_period"] = reference_data[date_column].dt.to_period(freq=freq)
                if current_data is not None:
                    current_data[date_column + "_period"] = current_data[date_column].dt.to_period(freq=freq)
                    feature_in_time_figure = self._plot_feature_in_time_2_df(
                        reference_data, current_data, date_column, feature_name, feature_type, color_options
                    )
                else:
                    feature_in_time_figure = self._plot_feature_in_time_1_df(
                        reference_data, date_column, feature_name, feature_type, color_options
                    )
                additional_graphs.append(
                    AdditionalGraphInfo(
                        feature_name + "_in_time",
                        {
                            "data": feature_in_time_figure["data"],
                            "layout": feature_in_time_figure["layout"],
                        },
                    )
                )

            if target_column and feature_name != target_column:
                if current_data is not None:
                    feature_and_target_figure = self._plot_feature_and_target_2_df(
                        reference_data,
                        current_data,
                        target_column,
                        target_type,
                        feature_name,
                        feature_type,
                        color_options,
                    )
                else:
                    feature_and_target_figure = self._plot_feature_and_target_1_df(
                        reference_data, target_column, target_type, feature_name, feature_type, color_options
                    )
                additional_graphs.append(
                    AdditionalGraphInfo(
                        feature_name + "_by_target",
                        {
                            "data": feature_and_target_figure["data"],
                            "layout": feature_and_target_figure["layout"],
                        },
                    )
                )

            wi = BaseWidgetInfo(
                type="rich_data",
                title="",
                size=2,
                params={
                    "header": feature_name,
                    "description": feature_type,
                    "metricsValuesHeaders": metrics_values_headers,
                    "metrics": self._metrics_for_table(feature_name, data_quality_results, is_current_data),
                    "graph": {"data": fig_main_distr["data"], "layout": fig_main_distr["layout"]},
                    "details": {"parts": parts, "insights": []},
                },
                additionalGraphs=additional_graphs,
            )

            widgets_list.append(wi)

        return BaseWidgetInfo(title="", size=2, type="list", widgets=widgets_list, pageSize=10)

    @staticmethod
    def _get_stats_with_names(
        stats_list: List[Tuple[str, str, Optional[str]]],
        reference_stats: FeatureQualityStats,
        current_stats: Optional[FeatureQualityStats],
    ) -> List[dict]:
        def get_values_as_string(stats_dict, field_name, field_percentage_name) -> str:
            field_value = stats_dict[field_name]

            if field_value is None:
                field_value = ""

            if field_percentage_name is None:
                return str(field_value)

            else:
                return f"{field_value} ({stats_dict[field_percentage_name]}%)"

        result = []
        reference_stats_dict = reference_stats.as_dict()

        if current_stats is None:
            current_stats_dict = None

        else:
            current_stats_dict = current_stats.as_dict()

        for stat_label, stat_field, stat_field_percentage in stats_list:
            values = [get_values_as_string(reference_stats_dict, stat_field, stat_field_percentage)]

            if current_stats_dict is not None:
                values.append(get_values_as_string(current_stats_dict, stat_field, stat_field_percentage))

            result.append(
                {
                    "label": stat_label,
                    "values": values,
                }
            )
        return result

    def _metrics_for_table(
        self, feature_name: str, data_quality_results: DataQualityAnalyzerResults, is_current_data: bool
    ):
        reference_stats = data_quality_results.reference_features_stats[feature_name]

        current_stats: Optional[FeatureQualityStats] = None

        if data_quality_results.current_features_stats is not None and is_current_data:
            current_stats = data_quality_results.current_features_stats[feature_name]

        metrics = []
        if reference_stats.is_category():
            # mapping for category stats: (field in analyser results name, label in widget table name)
            cat_features = [
                ("count", "count", None),
                ("unique", "unique_count", "unique_percentage"),
                ("most common", "most_common_value", "most_common_value_percentage"),
                ("missing", "missing_count", "missing_percentage"),
            ]

            if current_stats:
                cat_features.append(("new categories", "new_in_current_values_count", None))
                cat_features.append(("missing categories", "unused_in_current_values_count", None))

            metrics.extend(self._get_stats_with_names(cat_features, reference_stats, current_stats))

        elif reference_stats.is_numeric():
            # mapping for category stats: (label, field_name_for_main_value, field_name_for_percentage)
            num_features = [
                ("count", "count", None),
                ("mean", "mean", None),
                ("std", "std", None),
                ("min", "min", None),
                ("25%", "percentile_25", None),
                ("50%", "percentile_50", None),
                ("75%", "percentile_75", None),
                ("max", "max", None),
                ("unique", "unique_count", "unique_percentage"),
                ("most common", "most_common_value", "most_common_value_percentage"),
                ("missing", "missing_count", "missing_percentage"),
                ("infinite", "infinite_count", "infinite_percentage"),
            ]
            metrics.extend(self._get_stats_with_names(num_features, reference_stats, current_stats))

        elif reference_stats.is_datetime():
            # mapping for category stats: (field in analyser results name, label in widget table name)
            datetime_features = [
                ("count", "count", None),
                ("unique", "unique_count", "unique_percentage"),
                ("most common", "most_common_value", "most_common_value_percentage"),
                ("missing", "missing_count", "missing_percentage"),
                ("first", "min", None),
                ("last", "max", None),
            ]
            metrics.extend(self._get_stats_with_names(datetime_features, reference_stats, current_stats))

        return metrics

    def _plot_main_distr_figure(
        self,
        reference_data: pd.DataFrame,
        current_data: pd.DataFrame,
        feature_name: str,
        feature_type: str,
        color_options: ColorOptions,
    ) -> dict:
        if feature_type == "num":
            if current_data is None:
                trace1 = go.Histogram(x=reference_data[feature_name], marker_color=color_options.primary_color)
                trace2 = go.Histogram(
                    x=np.log10(reference_data.loc[reference_data[feature_name] > 0, feature_name]),
                    marker_color=color_options.primary_color,
                    visible=False,
                )
                data = [trace1, trace2]
                updatemenus = [
                    dict(
                        type="buttons",
                        direction="right",
                        x=1.0,
                        yanchor="top",
                        buttons=list(
                            [
                                dict(label="Linear Scale", method="update", args=[{"visible": [True, False]}]),
                                dict(label="Log Scale", method="update", args=[{"visible": [False, True]}]),
                            ]
                        ),
                    )
                ]

            else:
                trace1 = go.Histogram(
                    x=reference_data[feature_name],
                    marker_color=color_options.get_reference_data_color(),
                    name="reference",
                )
                trace2 = go.Histogram(
                    x=np.log10(reference_data.loc[reference_data[feature_name] > 0, feature_name]),
                    marker_color=color_options.get_reference_data_color(),
                    visible=False,
                    name="reference",
                )
                trace3 = go.Histogram(
                    x=current_data[feature_name], marker_color=color_options.get_current_data_color(), name="current"
                )
                trace4 = go.Histogram(
                    x=np.log10(current_data.loc[current_data[feature_name] > 0, feature_name]),
                    marker_color=color_options.get_current_data_color(),
                    visible=False,
                    name="current",
                )
                data = [trace1, trace2, trace3, trace4]

                updatemenus = [
                    dict(
                        type="buttons",
                        direction="right",
                        x=1.0,
                        yanchor="top",
                        buttons=list(
                            [
                                dict(
                                    label="Linear Scale",
                                    method="update",
                                    args=[{"visible": [True, False, True, False]}],
                                ),
                                dict(
                                    label="Log Scale", method="update", args=[{"visible": [False, True, False, True]}]
                                ),
                            ]
                        ),
                    )
                ]
            layout = dict(updatemenus=updatemenus)

            fig = go.Figure(data=data, layout=layout)

        elif feature_type == "cat":
            fig = go.Figure()
            cats = list(reference_data[feature_name].value_counts().index.astype(str))
            if "other" in cats:
                cats.remove("other")
                cats = cats + ["other"]
            if current_data is None:
                fig.add_trace(go.Histogram(x=reference_data[feature_name], marker_color=color_options.primary_color))
            else:
                fig.add_trace(
                    go.Histogram(
                        x=reference_data[feature_name],
                        marker_color=color_options.get_reference_data_color(),
                        name="reference",
                    )
                )
                fig.add_trace(
                    go.Histogram(
                        x=current_data[feature_name],
                        marker_color=color_options.get_current_data_color(),
                        name="current",
                    )
                )
            fig.update_xaxes(categoryorder="array", categoryarray=cats)

        elif feature_type == "datetime":
            freq = self._choose_agg_period(feature_name, reference_data, current_data)
            tmp_ref = reference_data[feature_name].dt.to_period(freq=freq)
            tmp_ref = tmp_ref.value_counts().reset_index()
            tmp_ref.columns = [feature_name, "number_of_items"]
            tmp_ref[feature_name] = tmp_ref[feature_name].dt.to_timestamp()
            fig = go.Figure()
            if current_data is None:
                fig.add_trace(
                    go.Scatter(
                        x=tmp_ref.sort_values(feature_name)[feature_name],
                        y=tmp_ref.sort_values(feature_name)["number_of_items"],
                        line=dict(color=color_options.primary_color, shape="spline"),
                    )
                )
            else:
                tmp_curr = current_data[feature_name].dt.to_period(freq=freq)
                tmp_curr = tmp_curr.value_counts().reset_index()
                tmp_curr.columns = [feature_name, "number_of_items"]
                tmp_curr[feature_name] = tmp_curr[feature_name].dt.to_timestamp()

                max_ref_date = tmp_ref[feature_name].max()
                min_curr_date = tmp_curr[feature_name].min()
                if max_ref_date == min_curr_date:
                    if (
                        tmp_curr.loc[tmp_curr[feature_name] == min_curr_date, "number_of_items"].iloc[0]
                        > tmp_ref.loc[tmp_ref[feature_name] == max_ref_date, "number_of_items"].iloc[0]
                    ):
                        tmp_curr.loc[tmp_curr[feature_name] == min_curr_date, "number_of_items"] = (
                            tmp_curr.loc[tmp_curr[feature_name] == min_curr_date, "number_of_items"]
                            + tmp_ref.loc[tmp_ref[feature_name] == max_ref_date, "number_of_items"]
                        )
                        tmp_ref = tmp_ref[tmp_ref[feature_name] != max_ref_date]
                    else:
                        tmp_ref.loc[tmp_ref[feature_name] == max_ref_date, "number_of_items"] = (
                            tmp_ref.loc[tmp_ref[feature_name] == max_ref_date, "number_of_items"]
                            + tmp_curr.loc[tmp_curr[feature_name] == min_curr_date, "number_of_items"]
                        )
                        tmp_curr = tmp_curr[tmp_curr[feature_name] != min_curr_date]

                fig.add_trace(
                    go.Scatter(
                        x=tmp_ref.sort_values(feature_name)[feature_name],
                        y=tmp_ref.sort_values(feature_name)["number_of_items"],
                        line=dict(color=color_options.get_reference_data_color(), shape="spline"),
                        name="reference",
                    )
                )
                fig.add_trace(
                    go.Scatter(
                        x=tmp_curr.sort_values(feature_name)[feature_name],
                        y=tmp_curr.sort_values(feature_name)["number_of_items"],
                        line=dict(color=color_options.get_current_data_color(), shape="spline"),
                        name="current",
                    )
                )
        else:
            return {}

        fig.update_layout(legend=dict(orientation="h", yanchor="bottom", y=1.02, xanchor="right", x=1))
        fig_main_distr = json.loads(fig.to_json())
        return fig_main_distr

    @staticmethod
    def assemble_parts(
        target_column: Optional[str], date_column: Optional[str], feature_name: str, feature_type: str
    ) -> List:
        parts = []
        if date_column and feature_type != "datetime":
            parts.append({"title": feature_name + " in time", "id": feature_name + "_in_time"})
        if target_column and feature_name != target_column:
            parts.append({"title": feature_name + " by target", "id": feature_name + "_by_target"})
        return parts

    @staticmethod
    def _transform_df_to_time_mean_view(df: pd.DataFrame, date_column: str, feature_name: str):
        df = df.groupby(date_column + "_period")[feature_name].mean().reset_index()
        df[date_column] = df[date_column + "_period"].dt.to_timestamp()
        return df

    @staticmethod
    def _transform_df_to_time_count_view(df: pd.DataFrame, date_column: str, feature_name: str):
        df = df.groupby([date_column + "_period", feature_name]).size()
        df.name = "num"
        df = df.reset_index()
        df[date_column] = df[date_column + "_period"].dt.to_timestamp()
        return df

    def _plot_feature_in_time_1_df(
        self,
        reference_data: pd.DataFrame,
        date_column: str,
        feature_name: str,
        feature_type: str,
        color_options: ColorOptions,
    ) -> dict:
        tmp = reference_data[[date_column + "_period", feature_name]].copy()
        feature_in_time_figure = {}

        if feature_type == "num":
            tmp = self._transform_df_to_time_mean_view(tmp, date_column, feature_name)

            fig = go.Figure()
            fig.add_trace(
                go.Scatter(
                    x=tmp.sort_values(date_column)[date_column],
                    y=tmp.sort_values(date_column)[feature_name],
                    line=dict(color=color_options.primary_color, shape="spline"),
                )
            )
            fig.update_layout(yaxis_title="Mean " + feature_name + " per " + self.period_prefix)
            feature_in_time_figure = json.loads(fig.to_json())

        elif feature_type == "cat":
            tmp = self._transform_df_to_time_count_view(tmp, date_column, feature_name)

            fig = go.Figure()
            for i, val in enumerate(tmp[feature_name].unique()):
                fig.add_trace(
                    go.Bar(
                        x=tmp.loc[tmp[feature_name] == val, date_column],
                        y=tmp.loc[tmp[feature_name] == val, "num"],
                        name=str(val),
                        marker_color=color_options.color_sequence[i],
                    )
                )
            fig.update_traces(marker_line_width=0.01)
            fig.update_layout(barmode="stack", bargap=0, yaxis_title="count category values per " + self.period_prefix)

            feature_in_time_figure = json.loads(fig.to_json())

        return feature_in_time_figure

    def _plot_feature_in_time_2_df(
        self,
        reference_data: pd.DataFrame,
        current_data: pd.DataFrame,
        date_column: str,
        feature_name: str,
        feature_type: str,
        color_options: ColorOptions,
    ) -> dict:
        tmp_ref = reference_data[[date_column + "_period", feature_name]].copy()
        tmp_curr = current_data[[date_column + "_period", feature_name]].copy()
        feature_in_time_figure = {}

        if feature_type == "num":
            tmp_ref = self._transform_df_to_time_mean_view(tmp_ref, date_column, feature_name)
            tmp_curr = self._transform_df_to_time_mean_view(tmp_curr, date_column, feature_name)

            fig = go.Figure()
            fig.add_trace(
                go.Scatter(
                    x=tmp_ref.sort_values(date_column)[date_column],
                    y=tmp_ref.sort_values(date_column)[feature_name],
                    line=dict(color=color_options.get_reference_data_color(), shape="spline"),
                    name="reference",
                )
            )
            fig.add_trace(
                go.Scatter(
                    x=tmp_curr.sort_values(date_column)[date_column],
                    y=tmp_curr.sort_values(date_column)[feature_name],
                    line=dict(color=color_options.get_current_data_color(), shape="spline"),
                    name="current",
                )
            )
            fig.update_layout(yaxis_title="Mean " + feature_name + " per " + self.period_prefix)
            feature_in_time_figure = json.loads(fig.to_json())

        elif feature_type == "cat":
            tmp_ref = self._transform_df_to_time_count_view(tmp_ref, date_column, feature_name)
            tmp_curr = self._transform_df_to_time_count_view(tmp_curr, date_column, feature_name)

            fig = go.Figure()
            for i, val in enumerate(tmp_ref[feature_name].unique()):
                fig.add_trace(
                    go.Bar(
                        x=tmp_ref.loc[tmp_ref[feature_name] == val, date_column],
                        y=tmp_ref.loc[tmp_ref[feature_name] == val, "num"],
                        name=str(val),
                        marker_color=color_options.color_sequence[i],
                        opacity=0.6,
                    )
                )
                fig.add_trace(
                    go.Bar(
                        x=tmp_curr.loc[tmp_curr[feature_name] == val, date_column],
                        y=tmp_curr.loc[tmp_curr[feature_name] == val, "num"],
                        name=str(val),
                        marker_color=color_options.color_sequence[i],
                        showlegend=False,
                    )
                )
            fig.update_traces(marker_line_width=0.01)
            fig.update_layout(
                barmode="stack",
                bargap=0,
                yaxis_title="count category values per " + self.period_prefix,
                title="reference/current",
            )
            feature_in_time_figure = json.loads(fig.to_json())

        return feature_in_time_figure

    @staticmethod
    def _get_subplots_ratio(s1: pd.Series, s2: pd.Series):
        d1 = (s1.max() - s1.min()).days
        d2 = (s2.max() - s2.min()).days
        ratio = np.round(d1 / (d1 + d2), 3)
        return ratio, 1 - ratio

    @staticmethod
    def _transform_df_count_values(df: pd.DataFrame, target_column: str, feature_name: Optional[str]):
        df = df.groupby([target_column, feature_name]).size()
        df.name = "count_objects"
        df = df.reset_index()
        return df

    def _plot_feature_and_target_1_df(
        self,
        reference_data: pd.DataFrame,
        target_column: str,
        target_type: Optional[str],
        feature_name: Optional[str],
        feature_type: str,
        color_options: ColorOptions,
    ) -> dict:
        tmp = reference_data[[target_column, feature_name]].copy()
        if feature_type == "cat":
            if target_type == "num":
                fig = go.Figure()
                trace = go.Box(x=tmp[feature_name], y=tmp[target_column], marker_color=color_options.primary_color)
                fig.add_trace(trace)
                fig.update_layout(yaxis_title=target_column, xaxis_title=feature_name)
                fig.update_traces(marker_size=3)
            else:
                tmp = self._transform_df_count_values(tmp, target_column, feature_name)
                fig = go.Figure()
                for i, val in enumerate(tmp[target_column].unique()):
                    trace = go.Bar(
                        x=tmp.loc[tmp[target_column] == val, feature_name],
                        y=tmp.loc[tmp[target_column] == val, "count_objects"],
                        marker_color=color_options.color_sequence[i],
                        name=str(val),
                    )
                    fig.add_trace(trace)
                fig.update_layout(yaxis_title="count")
        else:
            if target_type == "num":
                fig = go.Figure()
                trace = go.Scatter(
                    x=tmp.sample(min(2000, len(tmp)), random_state=0)[feature_name],
                    y=tmp.sample(min(2000, len(tmp)), random_state=0)[target_column],
                    mode="markers",
                    marker_color=color_options.primary_color,
                )
                fig.add_trace(trace)

                fig.update_layout(
                    yaxis_title=target_column, xaxis_title=feature_name, legend={"itemsizing": "constant"}
                )
                fig.update_traces(marker_size=4)

            else:
                fig = go.Figure()
                trace = go.Box(y=tmp[feature_name], x=tmp[target_column], marker_color=color_options.primary_color)
                fig.add_trace(trace)
                fig.update_layout(yaxis_title=feature_name, xaxis_title=target_column)
                fig.update_traces(marker_size=3)
        feature_and_target_figure = json.loads(fig.to_json())
        return feature_and_target_figure

    def _plot_feature_and_target_2_df(
        self,
        reference_data: pd.DataFrame,
        current_data: pd.DataFrame,
        target_column: str,
        target_type: Optional[str],
        feature_name: str,
        feature_type: str,
        color_options: ColorOptions,
    ) -> dict:
        tmp_ref = reference_data[[target_column, feature_name]].copy()
        tmp_curr = current_data[[target_column, feature_name]].copy()
        tmp_ref["df"] = "reference"
        tmp_curr["df"] = "current"
        if feature_type == "cat":
            if target_type == "num":
                fig = go.Figure()
                trace1 = go.Box(
                    x=tmp_ref[feature_name],
                    y=tmp_ref[target_column],
                    marker_color=color_options.get_reference_data_color(),
                    name="reference",
                )
                fig.add_trace(trace1)
                trace2 = go.Box(
                    x=tmp_curr[feature_name],
                    y=tmp_curr[target_column],
                    marker_color=color_options.get_current_data_color(),
                    name="current",
                )
                fig.add_trace(trace2)

                fig.update_layout(yaxis_title=target_column, xaxis_title=feature_name, boxmode="group")
                fig.update_traces(marker_size=3)
            else:
                tmp_ref = self._transform_df_count_values(tmp_ref, target_column, feature_name)
                tmp_curr = self._transform_df_count_values(tmp_curr, target_column, feature_name)
                fig = make_subplots(rows=1, cols=2, shared_yaxes=True, subplot_titles=["reference", "current"])

                for i, val in enumerate(tmp_ref[target_column].unique()):
                    trace = go.Bar(
                        x=tmp_ref.loc[tmp_ref[target_column] == val, feature_name],
                        y=tmp_ref.loc[tmp_ref[target_column] == val, "count_objects"],
                        marker_color=color_options.color_sequence[i],
                        opacity=0.6,
                        showlegend=False,
                    )
                    fig.add_trace(trace, 1, 1)
                for i, val in enumerate(tmp_curr[target_column].unique()):
                    trace = go.Bar(
                        x=tmp_curr.loc[tmp_curr[target_column] == val, feature_name],
                        y=tmp_curr.loc[tmp_curr[target_column] == val, "count_objects"],
                        marker_color=color_options.color_sequence[i],
                        name=str(val),
                    )
                    fig.add_trace(trace, 1, 2)
                fig.update_layout(yaxis_title="count")
        else:
            if target_type == "num":
                fig = make_subplots(rows=1, cols=2, shared_yaxes=True)
                trace = go.Scatter(
                    x=tmp_ref.sample(min(2000, len(tmp_ref)), random_state=0)[feature_name],
                    y=tmp_ref.sample(min(2000, len(tmp_ref)), random_state=0)[target_column],
                    mode="markers",
                    marker_color=color_options.get_reference_data_color(),
                    name="reference",
                )
                fig.add_trace(trace, 1, 1)
                trace = go.Scatter(
                    x=tmp_curr.sample(min(2000, len(tmp_curr)), random_state=0)[feature_name],
                    y=tmp_curr.sample(min(2000, len(tmp_curr)), random_state=0)[target_column],
                    mode="markers",
                    marker_color=color_options.get_current_data_color(),
                    name="current",
                )
                fig.add_trace(trace, 1, 2)
                fig.update_layout(yaxis_title=target_column, legend={"itemsizing": "constant"})
                fig.update_xaxes(title_text=feature_name, row=1, col=1)
                fig.update_xaxes(title_text=feature_name, row=1, col=2)
                fig.update_traces(marker_size=4)

            else:
                fig = go.Figure()
                trace1 = go.Box(
                    x=tmp_ref[target_column],
                    y=tmp_ref[feature_name],
                    marker_color=color_options.get_reference_data_color(),
                    name="reference",
                )
                fig.add_trace(trace1)
                trace2 = go.Box(
                    x=tmp_curr[target_column],
                    y=tmp_curr[feature_name],
                    marker_color=color_options.get_current_data_color(),
                    name="current",
                )
                fig.add_trace(trace2)
                fig.update_layout(yaxis_title=feature_name, xaxis_title=target_column, boxmode="group")
                fig.update_traces(marker_size=3)

        feature_and_target_figure = json.loads(fig.to_json())
        return feature_and_target_figure

    @staticmethod
    def _transform_cat_features(
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        cat_feature_names: List[str],
        target_column: Optional[str],
        target_type: Optional[str],
    ):
        if target_column and target_type == "cat":
            cat_feature_names = cat_feature_names + [target_column]
        for feature_name in cat_feature_names:
            if reference_data[feature_name].nunique() > 6:
                cats = reference_data[feature_name].value_counts().iloc[:5].index.astype(str)
                reference_data[feature_name] = reference_data[feature_name].apply(
                    lambda x: x if str(x) in cats else "other"
                )
                if current_data is not None:
                    current_data[feature_name] = current_data[feature_name].apply(
                        lambda x: x if str(x) in cats else "other"
                    )

    def _choose_agg_period(
        self, date_column: str, reference_data: pd.DataFrame, current_data: Optional[pd.DataFrame]
    ) -> str:
        optimal_points = 150
        prefix_dict = {"A": "year", "Q": "quarter", "M": "month", "W": "week", "D": "day", "H": "hour"}
        datetime_feature = reference_data[date_column]
        if current_data is not None:
            datetime_feature = datetime_feature.append(current_data[date_column])
        days = (datetime_feature.max() - datetime_feature.min()).days
        time_points = pd.Series(
            index=["A", "Q", "M", "W", "D", "H"],
            data=[
                abs(optimal_points - days / 365),
                abs(optimal_points - days / 90),
                abs(optimal_points - days / 30),
                abs(optimal_points - days / 7),
                abs(optimal_points - days),
                abs(optimal_points - days * 24),
            ],
        )
        self.period_prefix = prefix_dict[time_points.idxmin()]
        return str(time_points.idxmin())



import abc
from typing import Dict
from typing import Generic
from typing import List
from typing import Optional
from typing import Type
from typing import TypeVar
from typing import Union

from evidently.utils.data_operations import DatasetColumns

TObject = TypeVar("TObject")


class BaseGenerator(Generic[TObject]):
    """Base class for tests and metrics generator creation

    To create a new generator:
        - inherit a class from the base class
        - implement `generate_tests` method and return a list of test objects from it

    A Suite or a Report will call the method and add generated tests to its list instead of the generator object.

    You can use `columns_info` parameter in `generate` for getting data structure meta info like columns list.

    For example:
        if you want to create a test generator for 50, 90, 99 quantiles tests
        for all numeric columns with default condition, by reference quantiles
    >>> class TestQuantiles(BaseTestGenerator):
    ...    def generate(self, columns_info: DatasetColumns) -> List[TestValueQuantile]:
    ...        return [
    ...            TestColumnQuantile(column_name=name, quantile=quantile)
    ...            for quantile in (0.5, 0.9, 0.99)
    ...            for name in columns_info.num_feature_names
    ...        ]

    Do not forget set correct test type for `generate` return value
    """

    @abc.abstractmethod
    def generate(self, columns_info: DatasetColumns) -> List[TObject]:
        raise NotImplementedError()


def make_generator_by_columns(
    base_class: Type,
    columns: Optional[Union[str, list]] = None,
    parameters: Optional[Dict] = None,
    skip_id_column: bool = False,
) -> BaseGenerator:
    """Create a test generator for a columns list with a test class.

    Base class is specified with `base_class` parameter.
    If the test have no "column_name" parameter - TypeError will be raised.

    Columns list can be defined with parameter `columns`.
    If it is a list - just use it as a list of the columns.
    If `columns` is a string, it can be one of values:
    - "all" - make tests for all columns, including target/prediction columns
    - "num" - for numeric features
    - "cat" - for category features
    - "text" - for text features
    - "features" - for all features, not target/prediction columns.
    None value is the same as "all".
    If `columns` is string, and it is not one of the values, ValueError will be raised.

    `parameters` is used for specifying other parameters for each object, it is the same for all generated objects.
    """
    if parameters is None:
        parameters_for_generation: Dict = {}

    else:
        parameters_for_generation = parameters

    class ColumnsGenerator(BaseGenerator):
        def generate(self, columns_info: DatasetColumns) -> List[TObject]:
            nonlocal parameters_for_generation
            result = []

            if isinstance(columns, list):
                columns_for_generation = columns

            elif columns == "all" or columns is None:
                columns_for_generation = columns_info.get_all_columns_list(skip_id_column=skip_id_column)

            elif columns == "cat":
                columns_for_generation = columns_info.cat_feature_names

            elif columns == "num":
                columns_for_generation = columns_info.num_feature_names

            elif columns == "text":
                columns_for_generation = columns_info.text_feature_names

            elif columns == "features":
                columns_for_generation = columns_info.get_all_features_list(include_datetime_feature=True)

            else:
                raise ValueError("Incorrect parameter 'columns' for test generator")

            for column_name in columns_for_generation:
                parameters_for_generation["column_name"] = column_name
                # ignore possible parameters incompatibility
                # we cannot guarantee that a base class has column_name parameter
                # if it has not, type error will ve raised
                try:
                    result.append(base_class(**parameters_for_generation))  # type: ignore

                except TypeError as error:
                    raise TypeError(f"Cannot generate {base_class.__name__}. Error: {error}")

            return result

    return ColumnsGenerator()



#!/usr/bin/env python
# coding: utf-8

import json
from typing import Optional

import pandas as pd
import plotly.figure_factory as ff

from evidently import ColumnMapping
from evidently.analyzers.classification_performance_analyzer import (
    ClassificationPerformanceAnalyzer,
)
from evidently.dashboard.widgets.widget import Widget
from evidently.model.widget import BaseWidgetInfo


class ClassMetricsMatrixWidget(Widget):
    def __init__(self, title: str, dataset: str = "reference"):
        super().__init__(title)
        self.dataset = dataset  # reference or current

    def analyzers(self):
        return [ClassificationPerformanceAnalyzer]

    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results,
    ) -> Optional[BaseWidgetInfo]:

        results = ClassificationPerformanceAnalyzer.get_results(analyzers_results)
        target_name = results.columns.utility_columns.target
        prediction_name = results.columns.utility_columns.prediction

        if target_name is None or prediction_name is None:
            if self.dataset == "reference":
                raise ValueError(
                    f"Widget [{self.title}] requires 'target' and 'prediction' columns."
                )

            return None

        if self.dataset == "current":
            result_metrics = results.current_metrics

        elif self.dataset == "reference":
            result_metrics = results.reference_metrics

            if result_metrics is None:
                raise ValueError(
                    f"Widget [{self.title}] required 'reference' results from"
                    f" {ClassificationPerformanceAnalyzer.__name__} but no data found"
                )

        else:
            raise ValueError(
                f"Widget [{self.title}] requires 'current' or 'reference' dataset value"
            )

        if result_metrics is None:
            return None

        # plot support bar
        metrics_matrix = result_metrics.metrics_matrix
        metrics_frame = pd.DataFrame(metrics_matrix)

        z = metrics_frame.iloc[:-1, :-3].values

        x = (
            results.columns.target_names
            if results.columns.target_names
            else metrics_frame.columns.tolist()[:-3]
        )

        y = ["precision", "recall", "f1-score"]

        # change each element of z to type string for annotations
        z_text = [[str(round(y, 3)) for y in x] for x in z]

        # set up figure
        fig = ff.create_annotated_heatmap(
            z, x=x, y=y, annotation_text=z_text, colorscale="bluered", showscale=True
        )
        fig.update_layout(xaxis_title="Class", yaxis_title="Metric")

        metrics_matrix_json = json.loads(fig.to_json())

        return BaseWidgetInfo(
            title=self.title,
            type="big_graph",
            size=1 if current_data is not None else 2,
            params={
                "data": metrics_matrix_json["data"],
                "layout": metrics_matrix_json["layout"],
            },
        )



"""
Available metrics for Reports and Tests.
All metrics is grouped into modules.
For specific group see module documentation.
"""

from .classification_performance.class_balance_metric import ClassificationClassBalance
from .classification_performance.class_separation_metric import ClassificationClassSeparationPlot
from .classification_performance.classification_dummy_metric import ClassificationDummyMetric
from .classification_performance.classification_quality_metric import ClassificationQualityMetric
from .classification_performance.confusion_matrix_metric import ClassificationConfusionMatrix
from .classification_performance.pr_curve_metric import ClassificationPRCurve
from .classification_performance.pr_table_metric import ClassificationPRTable
from .classification_performance.probability_distribution_metric import ClassificationProbDistribution
from .classification_performance.quality_by_class_metric import ClassificationQualityByClass
from .classification_performance.quality_by_feature_table import ClassificationQualityByFeatureTable
from .classification_performance.roc_curve_metric import ClassificationRocCurve
from .data_drift.column_drift_metric import ColumnDriftMetric
from .data_drift.column_value_plot import ColumnValuePlot
from .data_drift.data_drift_table import DataDriftTable
from .data_drift.dataset_drift_metric import DatasetDriftMetric
from .data_drift.target_by_features_table import TargetByFeaturesTable
from .data_drift.text_descriptors_drift_metric import TextDescriptorsDriftMetric
from .data_integrity.column_missing_values_metric import ColumnMissingValuesMetric
from .data_integrity.column_regexp_metric import ColumnRegExpMetric
from .data_integrity.column_summary_metric import ColumnSummaryMetric
from .data_integrity.dataset_missing_values_metric import DatasetMissingValuesMetric
from .data_integrity.dataset_summary_metric import DatasetSummaryMetric
from .data_quality.column_correlations_metric import ColumnCorrelationsMetric
from .data_quality.column_distribution_metric import ColumnDistributionMetric
from .data_quality.column_quantile_metric import ColumnQuantileMetric
from .data_quality.column_value_list_metric import ColumnValueListMetric
from .data_quality.column_value_range_metric import ColumnValueRangeMetric
from .data_quality.conflict_prediction_metric import ConflictPredictionMetric
from .data_quality.conflict_target_metric import ConflictTargetMetric
from .data_quality.dataset_correlations_metric import DatasetCorrelationsMetric
from .data_quality.stability_metric import DataQualityStabilityMetric
from .data_quality.text_descriptors_correlation_metric import TextDescriptorsCorrelationMetric
from .data_quality.text_descriptors_distribution import TextDescriptorsDistribution
from .regression_performance.abs_perc_error_in_time import RegressionAbsPercentageErrorPlot
from .regression_performance.error_bias_table import RegressionErrorBiasTable
from .regression_performance.error_distribution import RegressionErrorDistribution
from .regression_performance.error_in_time import RegressionErrorPlot
from .regression_performance.error_normality import RegressionErrorNormality
from .regression_performance.predicted_and_actual_in_time import RegressionPredictedVsActualPlot
from .regression_performance.predicted_vs_actual import RegressionPredictedVsActualScatter
from .regression_performance.regression_dummy_metric import RegressionDummyMetric
from .regression_performance.regression_performance_metrics import RegressionPerformanceMetrics
from .regression_performance.regression_quality import RegressionQualityMetric
from .regression_performance.top_error import RegressionTopErrorMetric



#!/usr/bin/env python
# coding: utf-8

from dataclasses import dataclass
from typing import Optional

import pandas as pd

from evidently import ColumnMapping
from evidently.analyzers.base_analyzer import Analyzer
from evidently.analyzers.base_analyzer import BaseAnalyzerResult
from evidently.calculations.data_drift import DatasetDriftMetrics
from evidently.calculations.data_drift import get_drift_for_columns
from evidently.options import DataDriftOptions
from evidently.utils.data_operations import process_columns


@dataclass
class DataDriftAnalyzerResults(BaseAnalyzerResult):
    options: DataDriftOptions
    metrics: DatasetDriftMetrics


class DataDriftAnalyzer(Analyzer):
    @staticmethod
    def get_results(analyzer_results) -> DataDriftAnalyzerResults:
        return analyzer_results[DataDriftAnalyzer]

    def calculate(
        self, reference_data: pd.DataFrame, current_data: Optional[pd.DataFrame], column_mapping: ColumnMapping
    ) -> DataDriftAnalyzerResults:
        if current_data is None:
            raise ValueError("current_data should be present")

        data_drift_options = self.options_provider.get(DataDriftOptions)
        columns = process_columns(reference_data, column_mapping)
        result_metrics = get_drift_for_columns(
            current_data=current_data,
            reference_data=reference_data,
            dataset_columns=columns,
            data_drift_options=data_drift_options,
        )

        result = DataDriftAnalyzerResults(
            columns=columns,
            options=data_drift_options,
            metrics=result_metrics,
        )
        return result



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
import os
from os.path import join as pjoin

from setupbase import HERE
from setupbase import combine_commands
from setupbase import create_cmdclass
from setupbase import ensure_targets
from setupbase import install_npm
from setuptools import setup

nb_path = pjoin(HERE, "src", "evidently", "nbextension", "static")

# Representative files that should exist after a successful build
jstargets = [
    pjoin(nb_path, "index.js"),
]

package_data_spec = {
    "evidently": [
        "nbextension/static/*.*js*",
        "nbextension/static/*.*woff2*",
    ]
}

data_files_spec = [
    ("share/jupyter/nbextensions/evidently", nb_path, "*.js*"),
    ("share/jupyter/nbextensions/evidently", nb_path, "*.woff2"),
    ("etc/jupyter/nbconfig/notebook.d", HERE, "evidently.json"),
]

cmdclass = create_cmdclass("jsdeps", package_data_spec=package_data_spec, data_files_spec=data_files_spec)
cmdclass["jsdeps"] = combine_commands(
    install_npm(os.path.join(HERE, "ui"), build_cmd="build"),
    ensure_targets(jstargets),
)

setup_args = dict(
    cmdclass=cmdclass,
    author_email="emeli.dral@gmail.com",
    include_package_data=True,
    install_requires=[
        "plotly>=5.5.0",
        "statsmodels>=0.12.2",
        "scikit-learn>=0.24.0",
        "pandas>=1.3.5",
        "numpy>=1.19.5",
        "nltk>=3.6.7",
        "scipy>=1.5.4",
        "requests>=2.19.0",
        "PyYAML~=5.1",
    ],
    extras_require={
        "dev": [
            "wheel==0.35.1",
            "setuptools==50.3.2",
            "flake8==4.0.1",
            "jupyter==1.0.0",
            "mypy==0.910",
            "pytest==6.2.5",
            "types-PyYAML==6.0.1",
            "types-requests==2.26.0",
            "types-dataclasses==0.6",
            "black==22.8.0",
            "isort==5.10.1",
        ]
    },
    entry_points={},
)

if __name__ == "__main__":
    setup(**setup_args)



from abc import ABC
from typing import List
from typing import Optional
from typing import Union

from evidently.metrics import RegressionDummyMetric
from evidently.metrics import RegressionQualityMetric
from evidently.renderers.base_renderer import TestHtmlInfo
from evidently.renderers.base_renderer import TestRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import plotly_figure
from evidently.renderers.render_utils import plot_distr
from evidently.tests.base_test import BaseCheckValueTest
from evidently.tests.base_test import GroupData
from evidently.tests.base_test import GroupingTypes
from evidently.tests.base_test import TestValueCondition
from evidently.tests.utils import approx
from evidently.tests.utils import plot_check
from evidently.tests.utils import plot_metric_value
from evidently.tests.utils import regression_perf_plot
from evidently.utils.types import Numeric

REGRESSION_GROUP = GroupData("regression", "Regression", "")
GroupingTypes.TestGroup.add_value(REGRESSION_GROUP)


class BaseRegressionPerformanceMetricsTest(BaseCheckValueTest, ABC):
    group = REGRESSION_GROUP.id
    metric: RegressionQualityMetric
    dummy_metric: RegressionDummyMetric

    def __init__(
        self,
        eq: Optional[Numeric] = None,
        gt: Optional[Numeric] = None,
        gte: Optional[Numeric] = None,
        is_in: Optional[List[Union[Numeric, str, bool]]] = None,
        lt: Optional[Numeric] = None,
        lte: Optional[Numeric] = None,
        not_eq: Optional[Numeric] = None,
        not_in: Optional[List[Union[Numeric, str, bool]]] = None,
    ):
        super().__init__(eq=eq, gt=gt, gte=gte, is_in=is_in, lt=lt, lte=lte, not_eq=not_eq, not_in=not_in)
        self.metric = RegressionQualityMetric()
        self.dummy_metric = RegressionDummyMetric()


class TestValueMAE(BaseRegressionPerformanceMetricsTest):
    name = "Mean Absolute Error (MAE)"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition
        ref_mae = self.metric.get_result().mean_abs_error_ref
        if ref_mae is not None:
            return TestValueCondition(eq=approx(ref_mae, relative=0.1))
        return TestValueCondition(lt=self.dummy_metric.get_result().mean_abs_error_default)

    def calculate_value_for_test(self) -> Numeric:
        return self.metric.get_result().mean_abs_error

    def get_description(self, value: Numeric) -> str:
        return f"The MAE is {value:.3}. The test threshold is {self.get_condition()}"


@default_renderer(wrap_type=TestValueMAE)
class TestValueMAERenderer(TestRenderer):
    def render_json(self, obj: TestValueMAE) -> dict:
        base = super().render_json(obj)
        metric_result = obj.metric.get_result()
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["mean_abs_error"] = metric_result.mean_abs_error
        base["parameters"]["mean_abs_error_ref"] = metric_result.mean_abs_error_ref
        return base

    def render_html(self, obj: TestValueMAE) -> TestHtmlInfo:
        info = super().render_html(obj)
        is_ref_data = False

        if "reference" in obj.metric.get_result().hist_for_plot.keys():
            is_ref_data = True

        fig = regression_perf_plot(
            val_for_plot=obj.metric.get_result().vals_for_plots["mean_abs_error"],
            hist_for_plot=obj.metric.get_result().hist_for_plot,
            name="MAE",
            curr_metric=obj.metric.get_result().mean_abs_error,
            ref_metric=obj.metric.get_result().mean_abs_error_ref,
            is_ref_data=is_ref_data,
            color_options=self.color_options,
        )
        info.with_details("MAE", plotly_figure(title="", figure=fig))
        return info


class TestValueMAPE(BaseRegressionPerformanceMetricsTest):
    name = "Mean Absolute Percentage Error (MAPE)"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition
        ref_mae = self.metric.get_result().mean_abs_perc_error_ref
        if ref_mae is not None:
            return TestValueCondition(eq=approx(ref_mae, relative=0.1))
        return TestValueCondition(lt=self.dummy_metric.get_result().mean_abs_perc_error_default)

    def calculate_value_for_test(self) -> Numeric:
        return self.metric.get_result().mean_abs_perc_error

    def get_description(self, value: Numeric) -> str:
        return f"The MAPE is {value:.3}. The test threshold is {self.get_condition()}."


@default_renderer(wrap_type=TestValueMAPE)
class TestValueMAPERenderer(TestRenderer):
    def render_json(self, obj: TestValueMAPE) -> dict:
        base = super().render_json(obj)
        metric_result = obj.metric.get_result()
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["mean_abs_perc_error"] = metric_result.mean_abs_perc_error
        base["parameters"]["mean_abs_perc_error_ref"] = metric_result.mean_abs_perc_error_ref
        base["parameters"]["mean_abs_perc_error_default"] = metric_result.mean_abs_perc_error_default
        return base

    def render_html(self, obj: TestValueMAPE) -> TestHtmlInfo:
        info = super().render_html(obj)
        is_ref_data = False
        if "reference" in obj.metric.get_result().hist_for_plot.keys():
            is_ref_data = True
        val_for_plot = obj.metric.get_result().vals_for_plots["mean_abs_perc_error"]
        val_for_plot = {x: y * 100 for x, y in val_for_plot.items()}
        fig = regression_perf_plot(
            val_for_plot=val_for_plot,
            hist_for_plot=obj.metric.get_result().hist_for_plot,
            name="MAPE",
            curr_metric=obj.metric.get_result().mean_abs_perc_error,
            ref_metric=obj.metric.get_result().mean_abs_perc_error_ref,
            is_ref_data=is_ref_data,
            color_options=self.color_options,
        )
        info.with_details("MAPE", plotly_figure(title="", figure=fig))
        return info


class TestValueRMSE(BaseRegressionPerformanceMetricsTest):
    name = "Root Mean Square Error (RMSE)"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition
        rmse_ref = self.metric.get_result().rmse_ref
        if rmse_ref is not None:
            return TestValueCondition(eq=approx(rmse_ref, relative=0.1))
        return TestValueCondition(lt=self.dummy_metric.get_result().rmse_default)

    def calculate_value_for_test(self) -> Numeric:
        return self.metric.get_result().rmse

    def get_description(self, value: Numeric) -> str:
        return f"The RMSE is {value:.3}. The test threshold is {self.get_condition()}."


@default_renderer(wrap_type=TestValueRMSE)
class TestValueRMSERenderer(TestRenderer):
    def render_json(self, obj: TestValueRMSE) -> dict:
        base = super().render_json(obj)
        metric_result = obj.metric.get_result()
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["rmse"] = metric_result.rmse
        base["parameters"]["rmse_ref"] = metric_result.rmse_ref
        base["parameters"]["rmse_default"] = metric_result.rmse_default
        return base

    def render_html(self, obj: TestValueRMSE) -> TestHtmlInfo:
        info = super().render_html(obj)
        is_ref_data = False
        if "reference" in obj.metric.get_result().hist_for_plot.keys():
            is_ref_data = True
        fig = regression_perf_plot(
            val_for_plot=obj.metric.get_result().vals_for_plots["rmse"],
            hist_for_plot=obj.metric.get_result().hist_for_plot,
            name="RMSE",
            curr_metric=obj.metric.get_result().rmse,
            ref_metric=obj.metric.get_result().rmse_ref,
            is_ref_data=is_ref_data,
            color_options=self.color_options,
        )
        info.with_details("RMSE", plotly_figure(title="", figure=fig))
        return info


class TestValueMeanError(BaseRegressionPerformanceMetricsTest):
    name = "Mean Error (ME)"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition
        return TestValueCondition(eq=approx(0, absolute=0.1 * self.metric.get_result().me_default_sigma))

    def calculate_value_for_test(self) -> Numeric:
        return self.metric.get_result().mean_error

    def get_description(self, value: Numeric) -> str:
        return f"The ME is {value:.3}. The test threshold is {self.get_condition()}."


@default_renderer(wrap_type=TestValueMeanError)
class TestValueMeanErrorRenderer(TestRenderer):
    def render_json(self, obj: TestValueMeanError) -> dict:
        base = super().render_json(obj)
        metric_result = obj.metric.get_result()
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["mean_error"] = metric_result.mean_error
        return base

    def render_html(self, obj: TestValueMeanError) -> TestHtmlInfo:
        info = super().render_html(obj)
        me_hist_for_plot = obj.metric.get_result().me_hist_for_plot
        hist_curr = me_hist_for_plot["current"]
        hist_ref = None
        if "reference" in obj.metric.get_result().me_hist_for_plot.keys():
            hist_ref = me_hist_for_plot["reference"]
        fig = plot_distr(hist_curr=hist_curr, hist_ref=hist_ref, color_options=self.color_options)
        fig = plot_check(fig, obj.get_condition(), color_options=self.color_options)
        fig = plot_metric_value(fig, obj.metric.get_result().mean_error, "current mean error")
        info.with_details("", plotly_figure(title="", figure=fig))
        return info


class TestValueAbsMaxError(BaseRegressionPerformanceMetricsTest):
    name = "Max Absolute Error"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition
        abs_error_max_ref = self.metric.get_result().abs_error_max_ref
        if abs_error_max_ref is not None:
            return TestValueCondition(lte=approx(abs_error_max_ref, relative=0.1))
        return TestValueCondition(lte=self.dummy_metric.get_result().abs_error_max_default)

    def calculate_value_for_test(self) -> Numeric:
        return self.metric.get_result().abs_error_max

    def get_description(self, value: Numeric) -> str:
        return f"The Max Absolute Error is {value:.3}. The test threshold is {self.get_condition()}."


@default_renderer(wrap_type=TestValueAbsMaxError)
class TestValueAbsMaxErrorRenderer(TestRenderer):
    def render_json(self, obj: TestValueAbsMaxError) -> dict:
        base = super().render_json(obj)
        metric_result = obj.metric.get_result()
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["abs_error_max"] = metric_result.abs_error_max
        base["parameters"]["abs_error_max_ref"] = metric_result.abs_error_max_ref
        base["parameters"]["abs_error_max_ref"] = metric_result.abs_error_max_default
        return base

    def render_html(self, obj: TestValueAbsMaxError) -> TestHtmlInfo:
        info = super().render_html(obj)
        me_hist_for_plot = obj.metric.get_result().me_hist_for_plot
        hist_curr = me_hist_for_plot["current"]
        hist_ref = None

        if "reference" in obj.metric.get_result().me_hist_for_plot.keys():
            hist_ref = me_hist_for_plot["reference"]

        fig = plot_distr(hist_curr=hist_curr, hist_ref=hist_ref, color_options=self.color_options)
        info.with_details("", plotly_figure(title="", figure=fig))
        return info


class TestValueR2Score(BaseRegressionPerformanceMetricsTest):
    name = "R2 Score"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition
        r2_score_ref = self.metric.get_result().r2_score_ref
        if r2_score_ref is not None:
            return TestValueCondition(eq=approx(r2_score_ref, relative=0.1))
        return TestValueCondition(gt=0)

    def calculate_value_for_test(self) -> Numeric:
        return self.metric.get_result().r2_score

    def get_description(self, value: Numeric) -> str:
        return f"The R2 score is {value:.3}. The test threshold is {self.get_condition()}."


@default_renderer(wrap_type=TestValueR2Score)
class TestValueR2ScoreRenderer(TestRenderer):
    def render_json(self, obj: TestValueAbsMaxError) -> dict:
        base = super().render_json(obj)
        metric_result = obj.metric.get_result()
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["r2_score"] = metric_result.r2_score
        base["parameters"]["r2_score_ref"] = metric_result.r2_score_ref
        return base

    def render_html(self, obj: TestValueR2Score) -> TestHtmlInfo:
        info = super().render_html(obj)
        is_ref_data = False
        if "reference" in obj.metric.get_result().hist_for_plot.keys():
            is_ref_data = True
        fig = regression_perf_plot(
            val_for_plot=obj.metric.get_result().vals_for_plots["r2_score"],
            hist_for_plot=obj.metric.get_result().hist_for_plot,
            name="R2_score",
            curr_metric=obj.metric.get_result().r2_score,
            ref_metric=obj.metric.get_result().r2_score_ref,
            is_ref_data=is_ref_data,
            color_options=self.color_options,
        )
        info.with_details("R2 Score", plotly_figure(title="", figure=fig))
        return info
